// ----------------------------------------------------
// Brian Hamilton, 2015
//
// Description: code used for tests in Sections 4.1 and 4.2 of DAFx paper entitled 
//    "INVESTIGATING LARGE, COMPACT-IN-SPACE 3-D STENCIL OPERATIONS FOR ACOUSTICS SIMULATIONS ON A GPU"
// See paper for details.
// ----------------------------------------------------

//#-defined values
//precision: ReaL (float or double)
//stencil type: stype
//interior grid dims: Nx,Ny,Nz
//number of halo points: Nh
//number of points in stencil: K

//////////////////////////////////////////////////
//uncomment one of the following four groups
//in order to choose precision and grid size

   //#define ReaL double
   //#define Nx 672 //large grid
   //#define Ny 660 
   //#define Nz 600

   #define ReaL double
   #define Nx 640 //medium grid
   #define Ny 480
   #define Nz 420

   //#define ReaL float
   //#define Nx 928 //large grid
   //#define Ny 800
   //#define Nz 750 

   //#define ReaL float
   //#define Nx 720 //medium grid
   //#define Ny 640 
   //#define Nz 560 
//////////////////////////////////////////////////

#include "dafxheader.h" //this comes after #define ReaL

//////////////////////////////////////////////////
//uncomment one of the following groups
//to choose stencil with halo size and K

//compact-in-space stencils, R=stype
   #define stype 1
   #define Nh    1
   #define K     7

   //#define stype 2
   //#define Nh    1
   //#define K     19

   //#define stype 3
   //#define Nh    1
   //#define K     27

   //#define stype 4
   //#define Nh    2
   //#define K     33

   //#define stype 5
   //#define Nh    2
   //#define K     57

   //#define stype 6
   //#define Nh    2
   //#define K     81

   //#define stype 8
   //#define Nh    2
   //#define K     93

   //#define stype 9
   //#define Nh    3
   //#define K     123

   //#define stype 10
   //#define Nh    3
   //#define K     147

   //#define stype 11
   //#define Nh    3
   //#define K     171

   //#define stype 12
   //#define Nh    3
   //#define K     179

   //#define stype 13
   //#define Nh    3
   //#define K     203

   //#define stype 14
   //#define Nh    3
   //#define K     251

   //#define stype 16
   //#define Nh    4
   //#define K     257

   //#define stype 17
   //#define Nh    4
   //#define K     305

   //#define stype 18
   //#define Nh    4
   //#define K     341

   //#define stype 19
   //#define Nh    4
   //#define K     365

   //#define stype 20
   //#define Nh    4
   //#define K     389

   //#define stype 21
   //#define Nh    4
   //#define K     437

   //#define stype 22
   //#define Nh    4
   //#define K     461

//box-compact stencils, stype = last triplet in lexicographic ordering

   //#define stype 100
   //#define Nh    1
   //#define K     7

   //#define stype 110
   //#define Nh    1
   //#define K     19

   //#define stype 111
   //#define Nh    1
   //#define K     27

   //#define stype 200
   //#define Nh    2
   //#define K     33

   //#define stype 210
   //#define Nh    2
   //#define K     57

   //#define stype 211
   //#define Nh    2
   //#define K     81

   //#define stype 220
   //#define Nh    2
   //#define K     93

   //#define stype 221
   //#define Nh    2
   //#define K     117

   //#define stype 222
   //#define Nh    2
   //#define K     125

   //#define stype 300
   //#define Nh    3
   //#define K     131

   //#define stype 310
   //#define Nh    3
   //#define K     155

   //#define stype 311
   //#define Nh    3
   //#define K     179

   //#define stype 320
   //#define Nh    3
   //#define K     203

   //#define stype 321
   //#define Nh    3
   //#define K     251

   //#define stype 322
   //#define Nh    3
   //#define K     275

   //#define stype 330
   //#define Nh    3
   //#define K     287

   //#define stype 331
   //#define Nh    3
   //#define K     311

   //#define stype 332
   //#define Nh    3
   //#define K     335

   //#define stype 333
   //#define Nh    3
   //#define K     343

   //#define stype 400
   //#define Nh    4
   //#define K     349

//leggy stencils, stype = M*-100

   //#define stype -100
   //#define Nh    1
   //#define K     7

   //#define stype -200
   //#define Nh    2
   //#define K     13

   //#define stype -300
   //#define Nh    3
   //#define K     19

   //#define stype -400
   //#define Nh    4
   //#define K     25

   //#define stype -500
   //#define Nh    5
   //#define K     31

   //#define stype -600
   //#define Nh    6
   //#define K     37

   //#define stype -700
   //#define Nh    7
   //#define K     43

   //#define stype -800
   //#define Nh    8
   //#define K     49

   //#define stype -900
   //#define Nh    9
   //#define K     55

   //#define stype -1000
   //#define Nh    10
   //#define K     61

   //#define stype -1100
   //#define Nh    11
   //#define K     67

   //#define stype -1200
   //#define Nh    12
   //#define K     73

   //#define stype -1300
   //#define Nh    13
   //#define K     79

   //#define stype -1400
   //#define Nh    14
   //#define K     85

   //#define stype -1500
   //#define Nh    15
   //#define K     91

   //#define stype -1600
   //#define Nh    16
   //#define K     97

   //#define stype -1700
   //#define Nh    17
   //#define K     103

   //#define stype -1800
   //#define Nh    18
   //#define K     109

   //#define stype -1900
   //#define Nh    19
   //#define K     115

   //#define stype -2000
   //#define Nh    20
   //#define K     121
//////////////////////////////////////////////////

#define useKernelB 0 //use alternative kernel for leggy stencils (0 or 1)

//grid dims including halo
#define Nxh  (Nx+2*Nh)
#define Nyh  (Ny+2*Nh)
#define Nzh  (Nz+2*Nh)
#define area (Nxh*Nyh)

//read-out location
#define Rx 36
#define Ry 20
#define Rz 28

#define outputfname "room.bin"

// thread block sizes
#define Bx 32
#define By 4
#define Bz 2

#define maxNSh 23   //max number of shells
#define maxV 48     //max number of vectors per shell

__global__ void oout(ReaL *u0,ReaL *out,int n);
__global__ void KernelA(ReaL *u0, const ReaL * __restrict__ u1);
__global__ void KernelB(ReaL *u0, const ReaL * __restrict__ u1);

__constant__ ReaL gammaK[K];     //stencil coefficients
__constant__ int  offset[K];     //stencil vectors (offsets)

ReaL randomFloat()
{
   //random float between 1 and -1
   ReaL r = (ReaL)rand()/(ReaL)RAND_MAX*2.0-1.0;
   return r;
}

// ----------------------------------------------

int main(int argc, char *argv[]){

   printDevDetails();

   // ---------------------------------------------------------
   // Parameters
   // ---------------------------------------------------------
   int  N          = 100;     //number of samples to compute
   int  leggy      = (stype<0)? 1 : 0;

   // ---------------------------------------------------------
   // Setup
   // ---------------------------------------------------------
   ReaL gammaP[maxNSh+1];
   int  offsetx[maxNSh][maxV];
   int  offsety[maxNSh][maxV];
   int  offsetz[maxNSh][maxV];

   int  offset_h[K]; //stencil vectors (offsets)
   ReaL gammaK_h[K]; //stencil coefficients for KernelA

   srand(time(NULL));
   for (int qq=0;qq<maxNSh;qq++){
      gammaP[qq]=randomFloat();
   }

   //neighbours per shell
   int NN[maxNSh];
   for (int ii=0;ii<maxNSh;ii++){
      NN[ii]=0;
   }
   if (leggy==1) {
      for (int ii=0;ii<maxNSh;ii++){
         NN[ii]=6;
      }
   }
   //box-compact
   else if (stype > 99) {
      NN[0] = 6 ; //100
      NN[1] = 12; //110
      NN[2] = 8 ; //111
      NN[3] = 6 ; //200
      NN[4] = 24; //210
      NN[5] = 24; //211
      NN[6] = 12; //220
      NN[7] = 24; //221
      NN[8] = 8 ; //222
      NN[9] = 6 ; //300
      NN[10]= 24; //310
      NN[11]= 24; //311
      NN[12]= 24; //320
      NN[13]= 48; //321
      NN[14]= 24; //322
      NN[15]= 12; //330
      NN[16]= 24; //331
      NN[17]= 24; //332
      NN[18]= 8 ; //333
      NN[19]= 6 ; //400
   }
   //compact
   else { //squared norm - last shell
      NN[0] = 6 ; //1  - 100  
      NN[1] = 12; //2  - 110
      NN[2] = 8 ; //3  - 111
      NN[3] = 6 ; //4  - 200
      NN[4] = 24; //5  - 210
      NN[5] = 24; //6  - 211
      NN[6] = 12; //8  - 220
      NN[7] = 24; //9  - 221
      NN[8] = 6 ; //9  - 300
      NN[9] = 24; //10 - 310
      NN[10]= 24; //11 - 311
      NN[11]= 8 ; //12 - 222
      NN[12]= 24; //13 - 320
      NN[13]= 48; //14 - 321
      NN[14]= 6 ; //16 - 400
      NN[15]= 24; //17 - 322
      NN[16]= 24; //17 - 410
      NN[17]= 12; //18 - 330
      NN[18]= 24; //18 - 411
      NN[19]= 24; //19 - 331
      NN[20]= 24; //20 - 420
      NN[21]= 48; //21 - 421
      NN[22]= 24; //22 - 332
   }
   //number of shells (triplets) in stencil
   int NSh=0;
   if (leggy==1) {
      if      (stype== -100)  NSh=1 ;
      else if (stype== -200)  NSh=2 ;
      else if (stype== -300)  NSh=3 ;
      else if (stype== -400)  NSh=4 ;
      else if (stype== -500)  NSh=5 ;
      else if (stype== -600)  NSh=6 ;
      else if (stype== -700)  NSh=7 ;
      else if (stype== -800)  NSh=8 ;
      else if (stype== -900)  NSh=9 ;
      else if (stype==-1000)  NSh=10;
      else if (stype==-1100)  NSh=11;
      else if (stype==-1200)  NSh=12;
      else if (stype==-1300)  NSh=13;
      else if (stype==-1400)  NSh=14;
      else if (stype==-1500)  NSh=15;
      else if (stype==-1600)  NSh=16;
      else if (stype==-1700)  NSh=17;
      else if (stype==-1800)  NSh=18;
      else if (stype==-1900)  NSh=19;
      else if (stype==-2000)  NSh=20;
      else { 
         printf("invalid choice");exit(EXIT_FAILURE);
      }
   }
   //box-compact
   else if (stype > 99) {
      if      (stype== 100)  NSh=1 ;
      else if (stype== 110)  NSh=2 ;
      else if (stype== 111)  NSh=3 ;
      else if (stype== 200)  NSh=4 ;
      else if (stype== 210)  NSh=5 ;
      else if (stype== 211)  NSh=6 ;
      else if (stype== 220)  NSh=7 ;
      else if (stype== 221)  NSh=8 ;
      else if (stype== 222)  NSh=9 ;
      else if (stype== 300)  NSh=10;
      else if (stype== 310)  NSh=11;
      else if (stype== 311)  NSh=12;
      else if (stype== 320)  NSh=13;
      else if (stype== 321)  NSh=14;
      else if (stype== 322)  NSh=15;
      else if (stype== 330)  NSh=16;
      else if (stype== 331)  NSh=17;
      else if (stype== 332)  NSh=18;
      else if (stype== 333)  NSh=19;
      else if (stype== 400)  NSh=20;
      else { 
         printf("invalid choice");exit(EXIT_FAILURE);
      }
   }
   //compact
   else {
      if      (stype==1 )  NSh=1 ;
      else if (stype==2 )  NSh=2 ;
      else if (stype==3 )  NSh=3 ;
      else if (stype==4 )  NSh=4 ;
      else if (stype==5 )  NSh=5 ;
      else if (stype==6 )  NSh=6 ;
      else if (stype==8 )  NSh=7 ;
      else if (stype==9 )  NSh=9 ;
      else if (stype==10)  NSh=10;
      else if (stype==11)  NSh=11;
      else if (stype==12)  NSh=12;
      else if (stype==13)  NSh=13;
      else if (stype==14)  NSh=14;
      else if (stype==16)  NSh=15;
      else if (stype==17)  NSh=17;
      else if (stype==18)  NSh=19;
      else if (stype==19)  NSh=20;
      else if (stype==20)  NSh=21;
      else if (stype==21)  NSh=22;
      else if (stype==22)  NSh=23;
      else { 
         printf("invalid choice");exit(EXIT_FAILURE);
      }
   }

   //load offsets
   //start with zeros
   for (int ss=0;ss<maxNSh;ss++) {
      for (int qq=0;qq<maxV;qq++) {
         offsetx[ss][qq]=0; offsety[ss][qq]=0; offsetz[ss][qq]=0;
      }
   }
   //offsets for three dimensions
   //leggy
   if (leggy==1) {
      offsetx[0][0]=-1   ; offsety[0][0]=0  ; offsetz[0][0]=0  ; offsetx[0][1]=0  ; offsety[0][1]=-1   ; offsetz[0][1]=0  ; offsetx[0][2]=0  ; offsety[0][2]=0  ; offsetz[0][2]=-1   ; offsetx[0][3]=0  ; offsety[0][3]=0  ; offsetz[0][3]=1   ; offsetx[0][4]=0  ; offsety[0][4]=1   ; offsetz[0][4]=0  ; offsetx[0][5]=1   ; offsety[0][5]=0  ; offsetz[0][5]=0  ;
      offsetx[1][0]=-2   ; offsety[1][0]=0  ; offsetz[1][0]=0  ; offsetx[1][1]=0  ; offsety[1][1]=-2   ; offsetz[1][1]=0  ; offsetx[1][2]=0  ; offsety[1][2]=0  ; offsetz[1][2]=-2   ; offsetx[1][3]=0  ; offsety[1][3]=0  ; offsetz[1][3]=2   ; offsetx[1][4]=0  ; offsety[1][4]=2   ; offsetz[1][4]=0  ; offsetx[1][5]=2   ; offsety[1][5]=0  ; offsetz[1][5]=0  ;
      offsetx[2][0]=-3   ; offsety[2][0]=0  ; offsetz[2][0]=0  ; offsetx[2][1]=0  ; offsety[2][1]=-3   ; offsetz[2][1]=0  ; offsetx[2][2]=0  ; offsety[2][2]=0  ; offsetz[2][2]=-3   ; offsetx[2][3]=0  ; offsety[2][3]=0  ; offsetz[2][3]=3   ; offsetx[2][4]=0  ; offsety[2][4]=3   ; offsetz[2][4]=0  ; offsetx[2][5]=3   ; offsety[2][5]=0  ; offsetz[2][5]=0  ;
      offsetx[3][0]=-4   ; offsety[3][0]=0  ; offsetz[3][0]=0  ; offsetx[3][1]=0  ; offsety[3][1]=-4   ; offsetz[3][1]=0  ; offsetx[3][2]=0  ; offsety[3][2]=0  ; offsetz[3][2]=-4   ; offsetx[3][3]=0  ; offsety[3][3]=0  ; offsetz[3][3]=4   ; offsetx[3][4]=0  ; offsety[3][4]=4   ; offsetz[3][4]=0  ; offsetx[3][5]=4   ; offsety[3][5]=0  ; offsetz[3][5]=0  ;
      offsetx[4][0]=-5   ; offsety[4][0]=0  ; offsetz[4][0]=0  ; offsetx[4][1]=0  ; offsety[4][1]=-5   ; offsetz[4][1]=0  ; offsetx[4][2]=0  ; offsety[4][2]=0  ; offsetz[4][2]=-5   ; offsetx[4][3]=0  ; offsety[4][3]=0  ; offsetz[4][3]=5   ; offsetx[4][4]=0  ; offsety[4][4]=5   ; offsetz[4][4]=0  ; offsetx[4][5]=5   ; offsety[4][5]=0  ; offsetz[4][5]=0  ;
      offsetx[5][0]=-6   ; offsety[5][0]=0  ; offsetz[5][0]=0  ; offsetx[5][1]=0  ; offsety[5][1]=-6   ; offsetz[5][1]=0  ; offsetx[5][2]=0  ; offsety[5][2]=0  ; offsetz[5][2]=-6   ; offsetx[5][3]=0  ; offsety[5][3]=0  ; offsetz[5][3]=6   ; offsetx[5][4]=0  ; offsety[5][4]=6   ; offsetz[5][4]=0  ; offsetx[5][5]=6   ; offsety[5][5]=0  ; offsetz[5][5]=0  ;
      offsetx[6][0]=-7   ; offsety[6][0]=0  ; offsetz[6][0]=0  ; offsetx[6][1]=0  ; offsety[6][1]=-7   ; offsetz[6][1]=0  ; offsetx[6][2]=0  ; offsety[6][2]=0  ; offsetz[6][2]=-7   ; offsetx[6][3]=0  ; offsety[6][3]=0  ; offsetz[6][3]=7   ; offsetx[6][4]=0  ; offsety[6][4]=7   ; offsetz[6][4]=0  ; offsetx[6][5]=7   ; offsety[6][5]=0  ; offsetz[6][5]=0  ;
      offsetx[7][0]=-8   ; offsety[7][0]=0  ; offsetz[7][0]=0  ; offsetx[7][1]=0  ; offsety[7][1]=-8   ; offsetz[7][1]=0  ; offsetx[7][2]=0  ; offsety[7][2]=0  ; offsetz[7][2]=-8   ; offsetx[7][3]=0  ; offsety[7][3]=0  ; offsetz[7][3]=8   ; offsetx[7][4]=0  ; offsety[7][4]=8   ; offsetz[7][4]=0  ; offsetx[7][5]=8   ; offsety[7][5]=0  ; offsetz[7][5]=0  ;
      offsetx[8][0]=-9   ; offsety[8][0]=0  ; offsetz[8][0]=0  ; offsetx[8][1]=0  ; offsety[8][1]=-9   ; offsetz[8][1]=0  ; offsetx[8][2]=0  ; offsety[8][2]=0  ; offsetz[8][2]=-9   ; offsetx[8][3]=0  ; offsety[8][3]=0  ; offsetz[8][3]=9   ; offsetx[8][4]=0  ; offsety[8][4]=9   ; offsetz[8][4]=0  ; offsetx[8][5]=9   ; offsety[8][5]=0  ; offsetz[8][5]=0  ;
      offsetx[9][0]=-10  ; offsety[9][0]=0  ; offsetz[9][0]=0  ; offsetx[9][1]=0  ; offsety[9][1]=-10  ; offsetz[9][1]=0  ; offsetx[9][2]=0  ; offsety[9][2]=0  ; offsetz[9][2]=-10  ; offsetx[9][3]=0  ; offsety[9][3]=0  ; offsetz[9][3]=10  ; offsetx[9][4]=0  ; offsety[9][4]=10  ; offsetz[9][4]=0  ; offsetx[9][5]=10  ; offsety[9][5]=0  ; offsetz[9][5]=0  ;
      offsetx[10][0]=-11 ; offsety[10][0]=0 ; offsetz[10][0]=0 ; offsetx[10][1]=0 ; offsety[10][1]=-11 ; offsetz[10][1]=0 ; offsetx[10][2]=0 ; offsety[10][2]=0 ; offsetz[10][2]=-11 ; offsetx[10][3]=0 ; offsety[10][3]=0 ; offsetz[10][3]=11 ; offsetx[10][4]=0 ; offsety[10][4]=11 ; offsetz[10][4]=0 ; offsetx[10][5]=11 ; offsety[10][5]=0 ; offsetz[10][5]=0 ;
      offsetx[11][0]=-12 ; offsety[11][0]=0 ; offsetz[11][0]=0 ; offsetx[11][1]=0 ; offsety[11][1]=-12 ; offsetz[11][1]=0 ; offsetx[11][2]=0 ; offsety[11][2]=0 ; offsetz[11][2]=-12 ; offsetx[11][3]=0 ; offsety[11][3]=0 ; offsetz[11][3]=12 ; offsetx[11][4]=0 ; offsety[11][4]=12 ; offsetz[11][4]=0 ; offsetx[11][5]=12 ; offsety[11][5]=0 ; offsetz[11][5]=0 ;
      offsetx[12][0]=-13 ; offsety[12][0]=0 ; offsetz[12][0]=0 ; offsetx[12][1]=0 ; offsety[12][1]=-13 ; offsetz[12][1]=0 ; offsetx[12][2]=0 ; offsety[12][2]=0 ; offsetz[12][2]=-13 ; offsetx[12][3]=0 ; offsety[12][3]=0 ; offsetz[12][3]=13 ; offsetx[12][4]=0 ; offsety[12][4]=13 ; offsetz[12][4]=0 ; offsetx[12][5]=13 ; offsety[12][5]=0 ; offsetz[12][5]=0 ;
      offsetx[13][0]=-14 ; offsety[13][0]=0 ; offsetz[13][0]=0 ; offsetx[13][1]=0 ; offsety[13][1]=-14 ; offsetz[13][1]=0 ; offsetx[13][2]=0 ; offsety[13][2]=0 ; offsetz[13][2]=-14 ; offsetx[13][3]=0 ; offsety[13][3]=0 ; offsetz[13][3]=14 ; offsetx[13][4]=0 ; offsety[13][4]=14 ; offsetz[13][4]=0 ; offsetx[13][5]=14 ; offsety[13][5]=0 ; offsetz[13][5]=0 ;
      offsetx[14][0]=-15 ; offsety[14][0]=0 ; offsetz[14][0]=0 ; offsetx[14][1]=0 ; offsety[14][1]=-15 ; offsetz[14][1]=0 ; offsetx[14][2]=0 ; offsety[14][2]=0 ; offsetz[14][2]=-15 ; offsetx[14][3]=0 ; offsety[14][3]=0 ; offsetz[14][3]=15 ; offsetx[14][4]=0 ; offsety[14][4]=15 ; offsetz[14][4]=0 ; offsetx[14][5]=15 ; offsety[14][5]=0 ; offsetz[14][5]=0 ;
      offsetx[15][0]=-16 ; offsety[15][0]=0 ; offsetz[15][0]=0 ; offsetx[15][1]=0 ; offsety[15][1]=-16 ; offsetz[15][1]=0 ; offsetx[15][2]=0 ; offsety[15][2]=0 ; offsetz[15][2]=-16 ; offsetx[15][3]=0 ; offsety[15][3]=0 ; offsetz[15][3]=16 ; offsetx[15][4]=0 ; offsety[15][4]=16 ; offsetz[15][4]=0 ; offsetx[15][5]=16 ; offsety[15][5]=0 ; offsetz[15][5]=0 ;
      offsetx[16][0]=-17 ; offsety[16][0]=0 ; offsetz[16][0]=0 ; offsetx[16][1]=0 ; offsety[16][1]=-17 ; offsetz[16][1]=0 ; offsetx[16][2]=0 ; offsety[16][2]=0 ; offsetz[16][2]=-17 ; offsetx[16][3]=0 ; offsety[16][3]=0 ; offsetz[16][3]=17 ; offsetx[16][4]=0 ; offsety[16][4]=17 ; offsetz[16][4]=0 ; offsetx[16][5]=17 ; offsety[16][5]=0 ; offsetz[16][5]=0 ;
      offsetx[17][0]=-18 ; offsety[17][0]=0 ; offsetz[17][0]=0 ; offsetx[17][1]=0 ; offsety[17][1]=-18 ; offsetz[17][1]=0 ; offsetx[17][2]=0 ; offsety[17][2]=0 ; offsetz[17][2]=-18 ; offsetx[17][3]=0 ; offsety[17][3]=0 ; offsetz[17][3]=18 ; offsetx[17][4]=0 ; offsety[17][4]=18 ; offsetz[17][4]=0 ; offsetx[17][5]=18 ; offsety[17][5]=0 ; offsetz[17][5]=0 ;
      offsetx[18][0]=-19 ; offsety[18][0]=0 ; offsetz[18][0]=0 ; offsetx[18][1]=0 ; offsety[18][1]=-19 ; offsetz[18][1]=0 ; offsetx[18][2]=0 ; offsety[18][2]=0 ; offsetz[18][2]=-19 ; offsetx[18][3]=0 ; offsety[18][3]=0 ; offsetz[18][3]=19 ; offsetx[18][4]=0 ; offsety[18][4]=19 ; offsetz[18][4]=0 ; offsetx[18][5]=19 ; offsety[18][5]=0 ; offsetz[18][5]=0 ;
      offsetx[19][0]=-20 ; offsety[19][0]=0 ; offsetz[19][0]=0 ; offsetx[19][1]=0 ; offsety[19][1]=-20 ; offsetz[19][1]=0 ; offsetx[19][2]=0 ; offsety[19][2]=0 ; offsetz[19][2]=-20 ; offsetx[19][3]=0 ; offsety[19][3]=0 ; offsetz[19][3]=20 ; offsetx[19][4]=0 ; offsety[19][4]=20 ; offsetz[19][4]=0 ; offsetx[19][5]=20 ; offsety[19][5]=0 ; offsetz[19][5]=0 ;
   }
   //box-compact
   else if (stype>99){
      offsetx[0][0]=-1  ; offsety[0][0]=0   ; offsetz[0][0]=0   ; offsetx[0][1]=0   ; offsety[0][1]=-1  ; offsetz[0][1]=0   ; offsetx[0][2]=0   ; offsety[0][2]=0   ; offsetz[0][2]=-1  ; offsetx[0][3]=0   ; offsety[0][3]=0   ; offsetz[0][3]=1   ; offsetx[0][4]=0   ; offsety[0][4]=1   ; offsetz[0][4]=0   ; offsetx[0][5]=1   ; offsety[0][5]=0   ; offsetz[0][5]=0   ;
      offsetx[1][0]=-1  ; offsety[1][0]=-1  ; offsetz[1][0]=0   ; offsetx[1][1]=-1  ; offsety[1][1]=0   ; offsetz[1][1]=-1  ; offsetx[1][2]=-1  ; offsety[1][2]=0   ; offsetz[1][2]=1   ; offsetx[1][3]=-1  ; offsety[1][3]=1   ; offsetz[1][3]=0   ; offsetx[1][4]=0   ; offsety[1][4]=-1  ; offsetz[1][4]=-1  ; offsetx[1][5]=0   ; offsety[1][5]=-1  ; offsetz[1][5]=1   ; offsetx[1][6]=0   ; offsety[1][6]=1   ; offsetz[1][6]=-1  ; offsetx[1][7]=0   ; offsety[1][7]=1   ; offsetz[1][7]=1  ; offsetx[1][8]=1   ; offsety[1][8]=-1  ; offsetz[1][8]=0   ; offsetx[1][9]=1   ; offsety[1][9]=0   ; offsetz[1][9]=-1  ; offsetx[1][10]=1   ; offsety[1][10]=0   ; offsetz[1][10]=1   ; offsetx[1][11]=1   ; offsety[1][11]=1   ; offsetz[1][11]=0  ;
      offsetx[2][0]=-1  ; offsety[2][0]=-1  ; offsetz[2][0]=-1  ; offsetx[2][1]=-1  ; offsety[2][1]=-1  ; offsetz[2][1]=1   ; offsetx[2][2]=-1  ; offsety[2][2]=1   ; offsetz[2][2]=-1  ; offsetx[2][3]=-1  ; offsety[2][3]=1   ; offsetz[2][3]=1   ; offsetx[2][4]=1   ; offsety[2][4]=-1  ; offsetz[2][4]=-1  ; offsetx[2][5]=1   ; offsety[2][5]=-1  ; offsetz[2][5]=1   ; offsetx[2][6]=1   ; offsety[2][6]=1   ; offsetz[2][6]=-1  ; offsetx[2][7]=1   ; offsety[2][7]=1   ; offsetz[2][7]=1  ;
      offsetx[3][0]=-2  ; offsety[3][0]=0   ; offsetz[3][0]=0   ; offsetx[3][1]=0   ; offsety[3][1]=-2  ; offsetz[3][1]=0   ; offsetx[3][2]=0   ; offsety[3][2]=0   ; offsetz[3][2]=-2  ; offsetx[3][3]=0   ; offsety[3][3]=0   ; offsetz[3][3]=2   ; offsetx[3][4]=0   ; offsety[3][4]=2   ; offsetz[3][4]=0   ; offsetx[3][5]=2   ; offsety[3][5]=0   ; offsetz[3][5]=0   ;
      offsetx[4][0]=-2  ; offsety[4][0]=-1  ; offsetz[4][0]=0   ; offsetx[4][1]=-2  ; offsety[4][1]=0   ; offsetz[4][1]=-1  ; offsetx[4][2]=-2  ; offsety[4][2]=0   ; offsetz[4][2]=1   ; offsetx[4][3]=-2  ; offsety[4][3]=1   ; offsetz[4][3]=0   ; offsetx[4][4]=-1  ; offsety[4][4]=-2  ; offsetz[4][4]=0   ; offsetx[4][5]=-1  ; offsety[4][5]=0   ; offsetz[4][5]=-2  ; offsetx[4][6]=-1  ; offsety[4][6]=0   ; offsetz[4][6]=2   ; offsetx[4][7]=-1  ; offsety[4][7]=2   ; offsetz[4][7]=0  ; offsetx[4][8]=0   ; offsety[4][8]=-2  ; offsetz[4][8]=-1  ; offsetx[4][9]=0   ; offsety[4][9]=-2  ; offsetz[4][9]=1   ; offsetx[4][10]=0   ; offsety[4][10]=-1  ; offsetz[4][10]=-2  ; offsetx[4][11]=0   ; offsety[4][11]=-1  ; offsetz[4][11]=2  ; offsetx[4][12]=0   ; offsety[4][12]=1   ; offsetz[4][12]=-2  ; offsetx[4][13]=0   ; offsety[4][13]=1   ; offsetz[4][13]=2  ; offsetx[4][14]=0   ; offsety[4][14]=2   ; offsetz[4][14]=-1  ; offsetx[4][15]=0   ; offsety[4][15]=2   ; offsetz[4][15]=1  ; offsetx[4][16]=1   ; offsety[4][16]=-2  ; offsetz[4][16]=0   ; offsetx[4][17]=1   ; offsety[4][17]=0   ; offsetz[4][17]=-2  ; offsetx[4][18]=1   ; offsety[4][18]=0   ; offsetz[4][18]=2   ; offsetx[4][19]=1   ; offsety[4][19]=2   ; offsetz[4][19]=0  ; offsetx[4][20]=2   ; offsety[4][20]=-1  ; offsetz[4][20]=0   ; offsetx[4][21]=2   ; offsety[4][21]=0   ; offsetz[4][21]=-1  ; offsetx[4][22]=2   ; offsety[4][22]=0  ; offsetz[4][22]=1   ; offsetx[4][23]=2   ; offsety[4][23]=1  ; offsetz[4][23]=0  ;
      offsetx[5][0]=-2  ; offsety[5][0]=-1  ; offsetz[5][0]=-1  ; offsetx[5][1]=-2  ; offsety[5][1]=-1  ; offsetz[5][1]=1   ; offsetx[5][2]=-2  ; offsety[5][2]=1   ; offsetz[5][2]=-1  ; offsetx[5][3]=-2  ; offsety[5][3]=1   ; offsetz[5][3]=1   ; offsetx[5][4]=-1  ; offsety[5][4]=-2  ; offsetz[5][4]=-1  ; offsetx[5][5]=-1  ; offsety[5][5]=-2  ; offsetz[5][5]=1   ; offsetx[5][6]=-1  ; offsety[5][6]=-1  ; offsetz[5][6]=-2  ; offsetx[5][7]=-1  ; offsety[5][7]=-1  ; offsetz[5][7]=2  ; offsetx[5][8]=-1  ; offsety[5][8]=1   ; offsetz[5][8]=-2  ; offsetx[5][9]=-1  ; offsety[5][9]=1   ; offsetz[5][9]=2   ; offsetx[5][10]=-1  ; offsety[5][10]=2   ; offsetz[5][10]=-1  ; offsetx[5][11]=-1  ; offsety[5][11]=2   ; offsetz[5][11]=1  ; offsetx[5][12]=1   ; offsety[5][12]=-2  ; offsetz[5][12]=-1  ; offsetx[5][13]=1   ; offsety[5][13]=-2  ; offsetz[5][13]=1  ; offsetx[5][14]=1   ; offsety[5][14]=-1  ; offsetz[5][14]=-2  ; offsetx[5][15]=1   ; offsety[5][15]=-1  ; offsetz[5][15]=2  ; offsetx[5][16]=1   ; offsety[5][16]=1   ; offsetz[5][16]=-2  ; offsetx[5][17]=1   ; offsety[5][17]=1   ; offsetz[5][17]=2   ; offsetx[5][18]=1   ; offsety[5][18]=2   ; offsetz[5][18]=-1  ; offsetx[5][19]=1   ; offsety[5][19]=2   ; offsetz[5][19]=1  ; offsetx[5][20]=2   ; offsety[5][20]=-1  ; offsetz[5][20]=-1  ; offsetx[5][21]=2   ; offsety[5][21]=-1  ; offsetz[5][21]=1   ; offsetx[5][22]=2   ; offsety[5][22]=1  ; offsetz[5][22]=-1  ; offsetx[5][23]=2   ; offsety[5][23]=1  ; offsetz[5][23]=1  ;
      offsetx[6][0]=-2  ; offsety[6][0]=-2  ; offsetz[6][0]=0   ; offsetx[6][1]=-2  ; offsety[6][1]=0   ; offsetz[6][1]=-2  ; offsetx[6][2]=-2  ; offsety[6][2]=0   ; offsetz[6][2]=2   ; offsetx[6][3]=-2  ; offsety[6][3]=2   ; offsetz[6][3]=0   ; offsetx[6][4]=0   ; offsety[6][4]=-2  ; offsetz[6][4]=-2  ; offsetx[6][5]=0   ; offsety[6][5]=-2  ; offsetz[6][5]=2   ; offsetx[6][6]=0   ; offsety[6][6]=2   ; offsetz[6][6]=-2  ; offsetx[6][7]=0   ; offsety[6][7]=2   ; offsetz[6][7]=2  ; offsetx[6][8]=2   ; offsety[6][8]=-2  ; offsetz[6][8]=0   ; offsetx[6][9]=2   ; offsety[6][9]=0   ; offsetz[6][9]=-2  ; offsetx[6][10]=2   ; offsety[6][10]=0   ; offsetz[6][10]=2   ; offsetx[6][11]=2   ; offsety[6][11]=2   ; offsetz[6][11]=0  ;
      offsetx[7][0]=-2  ; offsety[7][0]=-2  ; offsetz[7][0]=-1  ; offsetx[7][1]=-2  ; offsety[7][1]=-2  ; offsetz[7][1]=1   ; offsetx[7][2]=-2  ; offsety[7][2]=-1  ; offsetz[7][2]=-2  ; offsetx[7][3]=-2  ; offsety[7][3]=-1  ; offsetz[7][3]=2   ; offsetx[7][4]=-2  ; offsety[7][4]=1   ; offsetz[7][4]=-2  ; offsetx[7][5]=-2  ; offsety[7][5]=1   ; offsetz[7][5]=2   ; offsetx[7][6]=-2  ; offsety[7][6]=2   ; offsetz[7][6]=-1  ; offsetx[7][7]=-2  ; offsety[7][7]=2   ; offsetz[7][7]=1  ; offsetx[7][8]=-1  ; offsety[7][8]=-2  ; offsetz[7][8]=-2  ; offsetx[7][9]=-1  ; offsety[7][9]=-2  ; offsetz[7][9]=2   ; offsetx[7][10]=-1  ; offsety[7][10]=2   ; offsetz[7][10]=-2  ; offsetx[7][11]=-1  ; offsety[7][11]=2   ; offsetz[7][11]=2  ; offsetx[7][12]=1   ; offsety[7][12]=-2  ; offsetz[7][12]=-2  ; offsetx[7][13]=1   ; offsety[7][13]=-2  ; offsetz[7][13]=2  ; offsetx[7][14]=1   ; offsety[7][14]=2   ; offsetz[7][14]=-2  ; offsetx[7][15]=1   ; offsety[7][15]=2   ; offsetz[7][15]=2  ; offsetx[7][16]=2   ; offsety[7][16]=-2  ; offsetz[7][16]=-1  ; offsetx[7][17]=2   ; offsety[7][17]=-2  ; offsetz[7][17]=1   ; offsetx[7][18]=2   ; offsety[7][18]=-1  ; offsetz[7][18]=-2  ; offsetx[7][19]=2   ; offsety[7][19]=-1  ; offsetz[7][19]=2  ; offsetx[7][20]=2   ; offsety[7][20]=1   ; offsetz[7][20]=-2  ; offsetx[7][21]=2   ; offsety[7][21]=1   ; offsetz[7][21]=2   ; offsetx[7][22]=2   ; offsety[7][22]=2  ; offsetz[7][22]=-1  ; offsetx[7][23]=2   ; offsety[7][23]=2  ; offsetz[7][23]=1  ;
      offsetx[8][0]=-2  ; offsety[8][0]=-2  ; offsetz[8][0]=-2  ; offsetx[8][1]=-2  ; offsety[8][1]=-2  ; offsetz[8][1]=2   ; offsetx[8][2]=-2  ; offsety[8][2]=2   ; offsetz[8][2]=-2  ; offsetx[8][3]=-2  ; offsety[8][3]=2   ; offsetz[8][3]=2   ; offsetx[8][4]=2   ; offsety[8][4]=-2  ; offsetz[8][4]=-2  ; offsetx[8][5]=2   ; offsety[8][5]=-2  ; offsetz[8][5]=2   ; offsetx[8][6]=2   ; offsety[8][6]=2   ; offsetz[8][6]=-2  ; offsetx[8][7]=2   ; offsety[8][7]=2   ; offsetz[8][7]=2  ;
      offsetx[9][0]=-3  ; offsety[9][0]=0   ; offsetz[9][0]=0   ; offsetx[9][1]=0   ; offsety[9][1]=-3  ; offsetz[9][1]=0   ; offsetx[9][2]=0   ; offsety[9][2]=0   ; offsetz[9][2]=-3  ; offsetx[9][3]=0   ; offsety[9][3]=0   ; offsetz[9][3]=3   ; offsetx[9][4]=0   ; offsety[9][4]=3   ; offsetz[9][4]=0   ; offsetx[9][5]=3   ; offsety[9][5]=0   ; offsetz[9][5]=0   ;
      offsetx[10][0]=-3 ; offsety[10][0]=-1 ; offsetz[10][0]=0  ; offsetx[10][1]=-3 ; offsety[10][1]=0  ; offsetz[10][1]=-1 ; offsetx[10][2]=-3 ; offsety[10][2]=0  ; offsetz[10][2]=1  ; offsetx[10][3]=-3 ; offsety[10][3]=1  ; offsetz[10][3]=0  ; offsetx[10][4]=-1 ; offsety[10][4]=-3 ; offsetz[10][4]=0  ; offsetx[10][5]=-1 ; offsety[10][5]=0  ; offsetz[10][5]=-3 ; offsetx[10][6]=-1 ; offsety[10][6]=0  ; offsetz[10][6]=3  ; offsetx[10][7]=-1 ; offsety[10][7]=3  ; offsetz[10][7]=0 ; offsetx[10][8]=0  ; offsety[10][8]=-3 ; offsetz[10][8]=-1 ; offsetx[10][9]=0  ; offsety[10][9]=-3 ; offsetz[10][9]=1  ; offsetx[10][10]=0  ; offsety[10][10]=-1 ; offsetz[10][10]=-3 ; offsetx[10][11]=0  ; offsety[10][11]=-1 ; offsetz[10][11]=3 ; offsetx[10][12]=0  ; offsety[10][12]=1  ; offsetz[10][12]=-3 ; offsetx[10][13]=0  ; offsety[10][13]=1  ; offsetz[10][13]=3 ; offsetx[10][14]=0  ; offsety[10][14]=3  ; offsetz[10][14]=-1 ; offsetx[10][15]=0  ; offsety[10][15]=3  ; offsetz[10][15]=1 ; offsetx[10][16]=1  ; offsety[10][16]=-3 ; offsetz[10][16]=0  ; offsetx[10][17]=1  ; offsety[10][17]=0  ; offsetz[10][17]=-3 ; offsetx[10][18]=1  ; offsety[10][18]=0  ; offsetz[10][18]=3  ; offsetx[10][19]=1  ; offsety[10][19]=3  ; offsetz[10][19]=0 ; offsetx[10][20]=3  ; offsety[10][20]=-1 ; offsetz[10][20]=0  ; offsetx[10][21]=3  ; offsety[10][21]=0  ; offsetz[10][21]=-1 ; offsetx[10][22]=3  ; offsety[10][22]=0 ; offsetz[10][22]=1  ; offsetx[10][23]=3  ; offsety[10][23]=1 ; offsetz[10][23]=0 ;
      offsetx[11][0]=-3 ; offsety[11][0]=-1 ; offsetz[11][0]=-1 ; offsetx[11][1]=-3 ; offsety[11][1]=-1 ; offsetz[11][1]=1  ; offsetx[11][2]=-3 ; offsety[11][2]=1  ; offsetz[11][2]=-1 ; offsetx[11][3]=-3 ; offsety[11][3]=1  ; offsetz[11][3]=1  ; offsetx[11][4]=-1 ; offsety[11][4]=-3 ; offsetz[11][4]=-1 ; offsetx[11][5]=-1 ; offsety[11][5]=-3 ; offsetz[11][5]=1  ; offsetx[11][6]=-1 ; offsety[11][6]=-1 ; offsetz[11][6]=-3 ; offsetx[11][7]=-1 ; offsety[11][7]=-1 ; offsetz[11][7]=3 ; offsetx[11][8]=-1 ; offsety[11][8]=1  ; offsetz[11][8]=-3 ; offsetx[11][9]=-1 ; offsety[11][9]=1  ; offsetz[11][9]=3  ; offsetx[11][10]=-1 ; offsety[11][10]=3  ; offsetz[11][10]=-1 ; offsetx[11][11]=-1 ; offsety[11][11]=3  ; offsetz[11][11]=1 ; offsetx[11][12]=1  ; offsety[11][12]=-3 ; offsetz[11][12]=-1 ; offsetx[11][13]=1  ; offsety[11][13]=-3 ; offsetz[11][13]=1 ; offsetx[11][14]=1  ; offsety[11][14]=-1 ; offsetz[11][14]=-3 ; offsetx[11][15]=1  ; offsety[11][15]=-1 ; offsetz[11][15]=3 ; offsetx[11][16]=1  ; offsety[11][16]=1  ; offsetz[11][16]=-3 ; offsetx[11][17]=1  ; offsety[11][17]=1  ; offsetz[11][17]=3  ; offsetx[11][18]=1  ; offsety[11][18]=3  ; offsetz[11][18]=-1 ; offsetx[11][19]=1  ; offsety[11][19]=3  ; offsetz[11][19]=1 ; offsetx[11][20]=3  ; offsety[11][20]=-1 ; offsetz[11][20]=-1 ; offsetx[11][21]=3  ; offsety[11][21]=-1 ; offsetz[11][21]=1  ; offsetx[11][22]=3  ; offsety[11][22]=1 ; offsetz[11][22]=-1 ; offsetx[11][23]=3  ; offsety[11][23]=1 ; offsetz[11][23]=1 ;
      offsetx[12][0]=-3 ; offsety[12][0]=-2 ; offsetz[12][0]=0  ; offsetx[12][1]=-3 ; offsety[12][1]=0  ; offsetz[12][1]=-2 ; offsetx[12][2]=-3 ; offsety[12][2]=0  ; offsetz[12][2]=2  ; offsetx[12][3]=-3 ; offsety[12][3]=2  ; offsetz[12][3]=0  ; offsetx[12][4]=-2 ; offsety[12][4]=-3 ; offsetz[12][4]=0  ; offsetx[12][5]=-2 ; offsety[12][5]=0  ; offsetz[12][5]=-3 ; offsetx[12][6]=-2 ; offsety[12][6]=0  ; offsetz[12][6]=3  ; offsetx[12][7]=-2 ; offsety[12][7]=3  ; offsetz[12][7]=0 ; offsetx[12][8]=0  ; offsety[12][8]=-3 ; offsetz[12][8]=-2 ; offsetx[12][9]=0  ; offsety[12][9]=-3 ; offsetz[12][9]=2  ; offsetx[12][10]=0  ; offsety[12][10]=-2 ; offsetz[12][10]=-3 ; offsetx[12][11]=0  ; offsety[12][11]=-2 ; offsetz[12][11]=3 ; offsetx[12][12]=0  ; offsety[12][12]=2  ; offsetz[12][12]=-3 ; offsetx[12][13]=0  ; offsety[12][13]=2  ; offsetz[12][13]=3 ; offsetx[12][14]=0  ; offsety[12][14]=3  ; offsetz[12][14]=-2 ; offsetx[12][15]=0  ; offsety[12][15]=3  ; offsetz[12][15]=2 ; offsetx[12][16]=2  ; offsety[12][16]=-3 ; offsetz[12][16]=0  ; offsetx[12][17]=2  ; offsety[12][17]=0  ; offsetz[12][17]=-3 ; offsetx[12][18]=2  ; offsety[12][18]=0  ; offsetz[12][18]=3  ; offsetx[12][19]=2  ; offsety[12][19]=3  ; offsetz[12][19]=0 ; offsetx[12][20]=3  ; offsety[12][20]=-2 ; offsetz[12][20]=0  ; offsetx[12][21]=3  ; offsety[12][21]=0  ; offsetz[12][21]=-2 ; offsetx[12][22]=3  ; offsety[12][22]=0 ; offsetz[12][22]=2  ; offsetx[12][23]=3  ; offsety[12][23]=2 ; offsetz[12][23]=0 ;
      offsetx[13][0]=-3 ; offsety[13][0]=-2 ; offsetz[13][0]=-1 ; offsetx[13][1]=-3 ; offsety[13][1]=-2 ; offsetz[13][1]=1  ; offsetx[13][2]=-3 ; offsety[13][2]=-1 ; offsetz[13][2]=-2 ; offsetx[13][3]=-3 ; offsety[13][3]=-1 ; offsetz[13][3]=2  ; offsetx[13][4]=-3 ; offsety[13][4]=1  ; offsetz[13][4]=-2 ; offsetx[13][5]=-3 ; offsety[13][5]=1  ; offsetz[13][5]=2  ; offsetx[13][6]=-3 ; offsety[13][6]=2  ; offsetz[13][6]=-1 ; offsetx[13][7]=-3 ; offsety[13][7]=2  ; offsetz[13][7]=1 ; offsetx[13][8]=-2 ; offsety[13][8]=-3 ; offsetz[13][8]=-1 ; offsetx[13][9]=-2 ; offsety[13][9]=-3 ; offsetz[13][9]=1  ; offsetx[13][10]=-2 ; offsety[13][10]=-1 ; offsetz[13][10]=-3 ; offsetx[13][11]=-2 ; offsety[13][11]=-1 ; offsetz[13][11]=3 ; offsetx[13][12]=-2 ; offsety[13][12]=1  ; offsetz[13][12]=-3 ; offsetx[13][13]=-2 ; offsety[13][13]=1  ; offsetz[13][13]=3 ; offsetx[13][14]=-2 ; offsety[13][14]=3  ; offsetz[13][14]=-1 ; offsetx[13][15]=-2 ; offsety[13][15]=3  ; offsetz[13][15]=1 ; offsetx[13][16]=-1 ; offsety[13][16]=-3 ; offsetz[13][16]=-2 ; offsetx[13][17]=-1 ; offsety[13][17]=-3 ; offsetz[13][17]=2  ; offsetx[13][18]=-1 ; offsety[13][18]=-2 ; offsetz[13][18]=-3 ; offsetx[13][19]=-1 ; offsety[13][19]=-2 ; offsetz[13][19]=3 ; offsetx[13][20]=-1 ; offsety[13][20]=2  ; offsetz[13][20]=-3 ; offsetx[13][21]=-1 ; offsety[13][21]=2  ; offsetz[13][21]=3  ; offsetx[13][22]=-1 ; offsety[13][22]=3 ; offsetz[13][22]=-2 ; offsetx[13][23]=-1 ; offsety[13][23]=3 ; offsetz[13][23]=2 ; offsetx[13][24]=1 ; offsety[13][24]=-3 ; offsetz[13][24]=-2 ; offsetx[13][25]=1 ; offsety[13][25]=-3 ; offsetz[13][25]=2 ; offsetx[13][26]=1 ; offsety[13][26]=-2 ; offsetz[13][26]=-3 ; offsetx[13][27]=1 ; offsety[13][27]=-2 ; offsetz[13][27]=3 ; offsetx[13][28]=1 ; offsety[13][28]=2 ; offsetz[13][28]=-3 ; offsetx[13][29]=1 ; offsety[13][29]=2 ; offsetz[13][29]=3 ; offsetx[13][30]=1 ; offsety[13][30]=3 ; offsetz[13][30]=-2 ; offsetx[13][31]=1 ; offsety[13][31]=3 ; offsetz[13][31]=2 ; offsetx[13][32]=2 ; offsety[13][32]=-3 ; offsetz[13][32]=-1 ; offsetx[13][33]=2 ; offsety[13][33]=-3 ; offsetz[13][33]=1 ; offsetx[13][34]=2 ; offsety[13][34]=-1 ; offsetz[13][34]=-3 ; offsetx[13][35]=2 ; offsety[13][35]=-1 ; offsetz[13][35]=3 ; offsetx[13][36]=2 ; offsety[13][36]=1 ; offsetz[13][36]=-3 ; offsetx[13][37]=2 ; offsety[13][37]=1 ; offsetz[13][37]=3 ; offsetx[13][38]=2 ; offsety[13][38]=3 ; offsetz[13][38]=-1 ; offsetx[13][39]=2 ; offsety[13][39]=3 ; offsetz[13][39]=1 ; offsetx[13][40]=3 ; offsety[13][40]=-2 ; offsetz[13][40]=-1 ; offsetx[13][41]=3 ; offsety[13][41]=-2 ; offsetz[13][41]=1 ; offsetx[13][42]=3 ; offsety[13][42]=-1 ; offsetz[13][42]=-2 ; offsetx[13][43]=3 ; offsety[13][43]=-1 ; offsetz[13][43]=2 ; offsetx[13][44]=3 ; offsety[13][44]=1 ; offsetz[13][44]=-2 ; offsetx[13][45]=3 ; offsety[13][45]=1 ; offsetz[13][45]=2 ; offsetx[13][46]=3 ; offsety[13][46]=2 ; offsetz[13][46]=-1 ; offsetx[13][47]=3 ; offsety[13][47]=2 ; offsetz[13][47]=1 ;
      offsetx[14][0]=-3 ; offsety[14][0]=-2 ; offsetz[14][0]=-2 ; offsetx[14][1]=-3 ; offsety[14][1]=-2 ; offsetz[14][1]=2  ; offsetx[14][2]=-3 ; offsety[14][2]=2  ; offsetz[14][2]=-2 ; offsetx[14][3]=-3 ; offsety[14][3]=2  ; offsetz[14][3]=2  ; offsetx[14][4]=-2 ; offsety[14][4]=-3 ; offsetz[14][4]=-2 ; offsetx[14][5]=-2 ; offsety[14][5]=-3 ; offsetz[14][5]=2  ; offsetx[14][6]=-2 ; offsety[14][6]=-2 ; offsetz[14][6]=-3 ; offsetx[14][7]=-2 ; offsety[14][7]=-2 ; offsetz[14][7]=3 ; offsetx[14][8]=-2 ; offsety[14][8]=2  ; offsetz[14][8]=-3 ; offsetx[14][9]=-2 ; offsety[14][9]=2  ; offsetz[14][9]=3  ; offsetx[14][10]=-2 ; offsety[14][10]=3  ; offsetz[14][10]=-2 ; offsetx[14][11]=-2 ; offsety[14][11]=3  ; offsetz[14][11]=2 ; offsetx[14][12]=2  ; offsety[14][12]=-3 ; offsetz[14][12]=-2 ; offsetx[14][13]=2  ; offsety[14][13]=-3 ; offsetz[14][13]=2 ; offsetx[14][14]=2  ; offsety[14][14]=-2 ; offsetz[14][14]=-3 ; offsetx[14][15]=2  ; offsety[14][15]=-2 ; offsetz[14][15]=3 ; offsetx[14][16]=2  ; offsety[14][16]=2  ; offsetz[14][16]=-3 ; offsetx[14][17]=2  ; offsety[14][17]=2  ; offsetz[14][17]=3  ; offsetx[14][18]=2  ; offsety[14][18]=3  ; offsetz[14][18]=-2 ; offsetx[14][19]=2  ; offsety[14][19]=3  ; offsetz[14][19]=2 ; offsetx[14][20]=3  ; offsety[14][20]=-2 ; offsetz[14][20]=-2 ; offsetx[14][21]=3  ; offsety[14][21]=-2 ; offsetz[14][21]=2  ; offsetx[14][22]=3  ; offsety[14][22]=2 ; offsetz[14][22]=-2 ; offsetx[14][23]=3  ; offsety[14][23]=2 ; offsetz[14][23]=2 ;
      offsetx[15][0]=-3 ; offsety[15][0]=-3 ; offsetz[15][0]=0  ; offsetx[15][1]=-3 ; offsety[15][1]=0  ; offsetz[15][1]=-3 ; offsetx[15][2]=-3 ; offsety[15][2]=0  ; offsetz[15][2]=3  ; offsetx[15][3]=-3 ; offsety[15][3]=3  ; offsetz[15][3]=0  ; offsetx[15][4]=0  ; offsety[15][4]=-3 ; offsetz[15][4]=-3 ; offsetx[15][5]=0  ; offsety[15][5]=-3 ; offsetz[15][5]=3  ; offsetx[15][6]=0  ; offsety[15][6]=3  ; offsetz[15][6]=-3 ; offsetx[15][7]=0  ; offsety[15][7]=3  ; offsetz[15][7]=3 ; offsetx[15][8]=3  ; offsety[15][8]=-3 ; offsetz[15][8]=0  ; offsetx[15][9]=3  ; offsety[15][9]=0  ; offsetz[15][9]=-3 ; offsetx[15][10]=3  ; offsety[15][10]=0  ; offsetz[15][10]=3  ; offsetx[15][11]=3  ; offsety[15][11]=3  ; offsetz[15][11]=0 ;
      offsetx[16][0]=-3 ; offsety[16][0]=-3 ; offsetz[16][0]=-1 ; offsetx[16][1]=-3 ; offsety[16][1]=-3 ; offsetz[16][1]=1  ; offsetx[16][2]=-3 ; offsety[16][2]=-1 ; offsetz[16][2]=-3 ; offsetx[16][3]=-3 ; offsety[16][3]=-1 ; offsetz[16][3]=3  ; offsetx[16][4]=-3 ; offsety[16][4]=1  ; offsetz[16][4]=-3 ; offsetx[16][5]=-3 ; offsety[16][5]=1  ; offsetz[16][5]=3  ; offsetx[16][6]=-3 ; offsety[16][6]=3  ; offsetz[16][6]=-1 ; offsetx[16][7]=-3 ; offsety[16][7]=3  ; offsetz[16][7]=1 ; offsetx[16][8]=-1 ; offsety[16][8]=-3 ; offsetz[16][8]=-3 ; offsetx[16][9]=-1 ; offsety[16][9]=-3 ; offsetz[16][9]=3  ; offsetx[16][10]=-1 ; offsety[16][10]=3  ; offsetz[16][10]=-3 ; offsetx[16][11]=-1 ; offsety[16][11]=3  ; offsetz[16][11]=3 ; offsetx[16][12]=1  ; offsety[16][12]=-3 ; offsetz[16][12]=-3 ; offsetx[16][13]=1  ; offsety[16][13]=-3 ; offsetz[16][13]=3 ; offsetx[16][14]=1  ; offsety[16][14]=3  ; offsetz[16][14]=-3 ; offsetx[16][15]=1  ; offsety[16][15]=3  ; offsetz[16][15]=3 ; offsetx[16][16]=3  ; offsety[16][16]=-3 ; offsetz[16][16]=-1 ; offsetx[16][17]=3  ; offsety[16][17]=-3 ; offsetz[16][17]=1  ; offsetx[16][18]=3  ; offsety[16][18]=-1 ; offsetz[16][18]=-3 ; offsetx[16][19]=3  ; offsety[16][19]=-1 ; offsetz[16][19]=3 ; offsetx[16][20]=3  ; offsety[16][20]=1  ; offsetz[16][20]=-3 ; offsetx[16][21]=3  ; offsety[16][21]=1  ; offsetz[16][21]=3  ; offsetx[16][22]=3  ; offsety[16][22]=3 ; offsetz[16][22]=-1 ; offsetx[16][23]=3  ; offsety[16][23]=3 ; offsetz[16][23]=1 ;
      offsetx[17][0]=-3 ; offsety[17][0]=-3 ; offsetz[17][0]=-2 ; offsetx[17][1]=-3 ; offsety[17][1]=-3 ; offsetz[17][1]=2  ; offsetx[17][2]=-3 ; offsety[17][2]=-2 ; offsetz[17][2]=-3 ; offsetx[17][3]=-3 ; offsety[17][3]=-2 ; offsetz[17][3]=3  ; offsetx[17][4]=-3 ; offsety[17][4]=2  ; offsetz[17][4]=-3 ; offsetx[17][5]=-3 ; offsety[17][5]=2  ; offsetz[17][5]=3  ; offsetx[17][6]=-3 ; offsety[17][6]=3  ; offsetz[17][6]=-2 ; offsetx[17][7]=-3 ; offsety[17][7]=3  ; offsetz[17][7]=2 ; offsetx[17][8]=-2 ; offsety[17][8]=-3 ; offsetz[17][8]=-3 ; offsetx[17][9]=-2 ; offsety[17][9]=-3 ; offsetz[17][9]=3  ; offsetx[17][10]=-2 ; offsety[17][10]=3  ; offsetz[17][10]=-3 ; offsetx[17][11]=-2 ; offsety[17][11]=3  ; offsetz[17][11]=3 ; offsetx[17][12]=2  ; offsety[17][12]=-3 ; offsetz[17][12]=-3 ; offsetx[17][13]=2  ; offsety[17][13]=-3 ; offsetz[17][13]=3 ; offsetx[17][14]=2  ; offsety[17][14]=3  ; offsetz[17][14]=-3 ; offsetx[17][15]=2  ; offsety[17][15]=3  ; offsetz[17][15]=3 ; offsetx[17][16]=3  ; offsety[17][16]=-3 ; offsetz[17][16]=-2 ; offsetx[17][17]=3  ; offsety[17][17]=-3 ; offsetz[17][17]=2  ; offsetx[17][18]=3  ; offsety[17][18]=-2 ; offsetz[17][18]=-3 ; offsetx[17][19]=3  ; offsety[17][19]=-2 ; offsetz[17][19]=3 ; offsetx[17][20]=3  ; offsety[17][20]=2  ; offsetz[17][20]=-3 ; offsetx[17][21]=3  ; offsety[17][21]=2  ; offsetz[17][21]=3  ; offsetx[17][22]=3  ; offsety[17][22]=3 ; offsetz[17][22]=-2 ; offsetx[17][23]=3  ; offsety[17][23]=3 ; offsetz[17][23]=2 ;
      offsetx[18][0]=-3 ; offsety[18][0]=-3 ; offsetz[18][0]=-3 ; offsetx[18][1]=-3 ; offsety[18][1]=-3 ; offsetz[18][1]=3  ; offsetx[18][2]=-3 ; offsety[18][2]=3  ; offsetz[18][2]=-3 ; offsetx[18][3]=-3 ; offsety[18][3]=3  ; offsetz[18][3]=3  ; offsetx[18][4]=3  ; offsety[18][4]=-3 ; offsetz[18][4]=-3 ; offsetx[18][5]=3  ; offsety[18][5]=-3 ; offsetz[18][5]=3  ; offsetx[18][6]=3  ; offsety[18][6]=3  ; offsetz[18][6]=-3 ; offsetx[18][7]=3  ; offsety[18][7]=3  ; offsetz[18][7]=3 ;
      offsetx[19][0]=-4 ; offsety[19][0]=0  ; offsetz[19][0]=0  ; offsetx[19][1]=0  ; offsety[19][1]=-4 ; offsetz[19][1]=0  ; offsetx[19][2]=0  ; offsety[19][2]=0  ; offsetz[19][2]=-4 ; offsetx[19][19]=0 ; offsety[19][19]=0 ; offsetz[19][19]=4 ; offsetx[19][4]=0  ; offsety[19][4]=4  ; offsetz[19][4]=0  ; offsetx[19][5]=4  ; offsety[19][5]=0  ; offsetz[19][5]=0  ;
   }
   //compact
   else {
      offsetx[0][0]=-1  ; offsety[0][0]=0   ; offsetz[0][0]=0   ; offsetx[0][1]=0   ; offsety[0][1]=-1  ; offsetz[0][1]=0   ; offsetx[0][2]=0   ; offsety[0][2]=0   ; offsetz[0][2]=-1  ; offsetx[0][3]=0   ; offsety[0][3]=0   ; offsetz[0][3]=1  ; offsetx[0][4]=0   ; offsety[0][4]=1   ; offsetz[0][4]=0   ; offsetx[0][5]=1   ; offsety[0][5]=0   ; offsetz[0][5]=0   ;
      offsetx[1][0]=-1  ; offsety[1][0]=-1  ; offsetz[1][0]=0   ; offsetx[1][1]=-1  ; offsety[1][1]=0   ; offsetz[1][1]=-1  ; offsetx[1][2]=-1  ; offsety[1][2]=0   ; offsetz[1][2]=1   ; offsetx[1][3]=-1  ; offsety[1][3]=1   ; offsetz[1][3]=0  ; offsetx[1][4]=0   ; offsety[1][4]=-1  ; offsetz[1][4]=-1  ; offsetx[1][5]=0   ; offsety[1][5]=-1  ; offsetz[1][5]=1   ; offsetx[1][6]=0   ; offsety[1][6]=1   ; offsetz[1][6]=-1  ; offsetx[1][7]=0   ; offsety[1][7]=1   ; offsetz[1][7]=1  ; offsetx[1][8]=1   ; offsety[1][8]=-1  ; offsetz[1][8]=0   ; offsetx[1][9]=1   ; offsety[1][9]=0   ; offsetz[1][9]=-1  ; offsetx[1][10]=1   ; offsety[1][10]=0   ; offsetz[1][10]=1   ; offsetx[1][11]=1   ; offsety[1][11]=1   ; offsetz[1][11]=0  ;
      offsetx[2][0]=-1  ; offsety[2][0]=-1  ; offsetz[2][0]=-1  ; offsetx[2][1]=-1  ; offsety[2][1]=-1  ; offsetz[2][1]=1   ; offsetx[2][2]=-1  ; offsety[2][2]=1   ; offsetz[2][2]=-1  ; offsetx[2][3]=-1  ; offsety[2][3]=1   ; offsetz[2][3]=1  ; offsetx[2][4]=1   ; offsety[2][4]=-1  ; offsetz[2][4]=-1  ; offsetx[2][5]=1   ; offsety[2][5]=-1  ; offsetz[2][5]=1   ; offsetx[2][6]=1   ; offsety[2][6]=1   ; offsetz[2][6]=-1  ; offsetx[2][7]=1   ; offsety[2][7]=1   ; offsetz[2][7]=1  ;
      offsetx[3][0]=-2  ; offsety[3][0]=0   ; offsetz[3][0]=0   ; offsetx[3][1]=0   ; offsety[3][1]=-2  ; offsetz[3][1]=0   ; offsetx[3][2]=0   ; offsety[3][2]=0   ; offsetz[3][2]=-2  ; offsetx[3][3]=0   ; offsety[3][3]=0   ; offsetz[3][3]=2  ; offsetx[3][4]=0   ; offsety[3][4]=2   ; offsetz[3][4]=0   ; offsetx[3][5]=2   ; offsety[3][5]=0   ; offsetz[3][5]=0   ;
      offsetx[4][0]=-2  ; offsety[4][0]=-1  ; offsetz[4][0]=0   ; offsetx[4][1]=-2  ; offsety[4][1]=0   ; offsetz[4][1]=-1  ; offsetx[4][2]=-2  ; offsety[4][2]=0   ; offsetz[4][2]=1   ; offsetx[4][3]=-2  ; offsety[4][3]=1   ; offsetz[4][3]=0  ; offsetx[4][4]=-1  ; offsety[4][4]=-2  ; offsetz[4][4]=0   ; offsetx[4][5]=-1  ; offsety[4][5]=0   ; offsetz[4][5]=-2  ; offsetx[4][6]=-1  ; offsety[4][6]=0   ; offsetz[4][6]=2   ; offsetx[4][7]=-1  ; offsety[4][7]=2   ; offsetz[4][7]=0  ; offsetx[4][8]=0   ; offsety[4][8]=-2  ; offsetz[4][8]=-1  ; offsetx[4][9]=0   ; offsety[4][9]=-2  ; offsetz[4][9]=1   ; offsetx[4][10]=0   ; offsety[4][10]=-1  ; offsetz[4][10]=-2  ; offsetx[4][11]=0   ; offsety[4][11]=-1  ; offsetz[4][11]=2  ; offsetx[4][12]=0   ; offsety[4][12]=1   ; offsetz[4][12]=-2  ; offsetx[4][13]=0   ; offsety[4][13]=1   ; offsetz[4][13]=2  ; offsetx[4][14]=0   ; offsety[4][14]=2   ; offsetz[4][14]=-1  ; offsetx[4][15]=0   ; offsety[4][15]=2   ; offsetz[4][15]=1  ; offsetx[4][16]=1   ; offsety[4][16]=-2  ; offsetz[4][16]=0   ; offsetx[4][17]=1   ; offsety[4][17]=0   ; offsetz[4][17]=-2  ; offsetx[4][18]=1   ; offsety[4][18]=0   ; offsetz[4][18]=2   ; offsetx[4][19]=1   ; offsety[4][19]=2   ; offsetz[4][19]=0  ; offsetx[4][20]=2   ; offsety[4][20]=-1  ; offsetz[4][20]=0   ; offsetx[4][21]=2   ; offsety[4][21]=0   ; offsetz[4][21]=-1  ; offsetx[4][22]=2   ; offsety[4][22]=0  ; offsetz[4][22]=1   ; offsetx[4][23]=2   ; offsety[4][23]=1  ; offsetz[4][23]=0  ;
      offsetx[5][0]=-2  ; offsety[5][0]=-1  ; offsetz[5][0]=-1  ; offsetx[5][1]=-2  ; offsety[5][1]=-1  ; offsetz[5][1]=1   ; offsetx[5][2]=-2  ; offsety[5][2]=1   ; offsetz[5][2]=-1  ; offsetx[5][3]=-2  ; offsety[5][3]=1   ; offsetz[5][3]=1  ; offsetx[5][4]=-1  ; offsety[5][4]=-2  ; offsetz[5][4]=-1  ; offsetx[5][5]=-1  ; offsety[5][5]=-2  ; offsetz[5][5]=1   ; offsetx[5][6]=-1  ; offsety[5][6]=-1  ; offsetz[5][6]=-2  ; offsetx[5][7]=-1  ; offsety[5][7]=-1  ; offsetz[5][7]=2  ; offsetx[5][8]=-1  ; offsety[5][8]=1   ; offsetz[5][8]=-2  ; offsetx[5][9]=-1  ; offsety[5][9]=1   ; offsetz[5][9]=2   ; offsetx[5][10]=-1  ; offsety[5][10]=2   ; offsetz[5][10]=-1  ; offsetx[5][11]=-1  ; offsety[5][11]=2   ; offsetz[5][11]=1  ; offsetx[5][12]=1   ; offsety[5][12]=-2  ; offsetz[5][12]=-1  ; offsetx[5][13]=1   ; offsety[5][13]=-2  ; offsetz[5][13]=1  ; offsetx[5][14]=1   ; offsety[5][14]=-1  ; offsetz[5][14]=-2  ; offsetx[5][15]=1   ; offsety[5][15]=-1  ; offsetz[5][15]=2  ; offsetx[5][16]=1   ; offsety[5][16]=1   ; offsetz[5][16]=-2  ; offsetx[5][17]=1   ; offsety[5][17]=1   ; offsetz[5][17]=2   ; offsetx[5][18]=1   ; offsety[5][18]=2   ; offsetz[5][18]=-1  ; offsetx[5][19]=1   ; offsety[5][19]=2   ; offsetz[5][19]=1  ; offsetx[5][20]=2   ; offsety[5][20]=-1  ; offsetz[5][20]=-1  ; offsetx[5][21]=2   ; offsety[5][21]=-1  ; offsetz[5][21]=1   ; offsetx[5][22]=2   ; offsety[5][22]=1  ; offsetz[5][22]=-1  ; offsetx[5][23]=2   ; offsety[5][23]=1  ; offsetz[5][23]=1  ;
      offsetx[6][0]=-2  ; offsety[6][0]=-2  ; offsetz[6][0]=0   ; offsetx[6][1]=-2  ; offsety[6][1]=0   ; offsetz[6][1]=-2  ; offsetx[6][2]=-2  ; offsety[6][2]=0   ; offsetz[6][2]=2   ; offsetx[6][3]=-2  ; offsety[6][3]=2   ; offsetz[6][3]=0  ; offsetx[6][4]=0   ; offsety[6][4]=-2  ; offsetz[6][4]=-2  ; offsetx[6][5]=0   ; offsety[6][5]=-2  ; offsetz[6][5]=2   ; offsetx[6][6]=0   ; offsety[6][6]=2   ; offsetz[6][6]=-2  ; offsetx[6][7]=0   ; offsety[6][7]=2   ; offsetz[6][7]=2  ; offsetx[6][8]=2   ; offsety[6][8]=-2  ; offsetz[6][8]=0   ; offsetx[6][9]=2   ; offsety[6][9]=0   ; offsetz[6][9]=-2  ; offsetx[6][10]=2   ; offsety[6][10]=0   ; offsetz[6][10]=2   ; offsetx[6][11]=2   ; offsety[6][11]=2   ; offsetz[6][11]=0  ;
      offsetx[7][0]=-2  ; offsety[7][0]=-2  ; offsetz[7][0]=-1  ; offsetx[7][1]=-2  ; offsety[7][1]=-2  ; offsetz[7][1]=1   ; offsetx[7][2]=-2  ; offsety[7][2]=-1  ; offsetz[7][2]=-2  ; offsetx[7][3]=-2  ; offsety[7][3]=-1  ; offsetz[7][3]=2  ; offsetx[7][4]=-2  ; offsety[7][4]=1   ; offsetz[7][4]=-2  ; offsetx[7][5]=-2  ; offsety[7][5]=1   ; offsetz[7][5]=2   ; offsetx[7][6]=-2  ; offsety[7][6]=2   ; offsetz[7][6]=-1  ; offsetx[7][7]=-2  ; offsety[7][7]=2   ; offsetz[7][7]=1  ; offsetx[7][8]=-1  ; offsety[7][8]=-2  ; offsetz[7][8]=-2  ; offsetx[7][9]=-1  ; offsety[7][9]=-2  ; offsetz[7][9]=2   ; offsetx[7][10]=-1  ; offsety[7][10]=2   ; offsetz[7][10]=-2  ; offsetx[7][11]=-1  ; offsety[7][11]=2   ; offsetz[7][11]=2  ; offsetx[7][12]=1   ; offsety[7][12]=-2  ; offsetz[7][12]=-2  ; offsetx[7][13]=1   ; offsety[7][13]=-2  ; offsetz[7][13]=2  ; offsetx[7][14]=1   ; offsety[7][14]=2   ; offsetz[7][14]=-2  ; offsetx[7][15]=1   ; offsety[7][15]=2   ; offsetz[7][15]=2  ; offsetx[7][16]=2   ; offsety[7][16]=-2  ; offsetz[7][16]=-1  ; offsetx[7][17]=2   ; offsety[7][17]=-2  ; offsetz[7][17]=1   ; offsetx[7][18]=2   ; offsety[7][18]=-1  ; offsetz[7][18]=-2  ; offsetx[7][19]=2   ; offsety[7][19]=-1  ; offsetz[7][19]=2  ; offsetx[7][20]=2   ; offsety[7][20]=1   ; offsetz[7][20]=-2  ; offsetx[7][21]=2   ; offsety[7][21]=1   ; offsetz[7][21]=2   ; offsetx[7][22]=2   ; offsety[7][22]=2  ; offsetz[7][22]=-1  ; offsetx[7][23]=2   ; offsety[7][23]=2  ; offsetz[7][23]=1  ;
      offsetx[8][0]=-3  ; offsety[8][0]=0   ; offsetz[8][0]=0   ; offsetx[8][1]=0   ; offsety[8][1]=-3  ; offsetz[8][1]=0   ; offsetx[8][2]=0   ; offsety[8][2]=0   ; offsetz[8][2]=-3  ; offsetx[8][3]=0   ; offsety[8][3]=0   ; offsetz[8][3]=3  ; offsetx[8][4]=0   ; offsety[8][4]=3   ; offsetz[8][4]=0   ; offsetx[8][5]=3   ; offsety[8][5]=0   ; offsetz[8][5]=0   ;
      offsetx[9][0]=-3  ; offsety[9][0]=-1  ; offsetz[9][0]=0   ; offsetx[9][1]=-3  ; offsety[9][1]=0   ; offsetz[9][1]=-1  ; offsetx[9][2]=-3  ; offsety[9][2]=0   ; offsetz[9][2]=1   ; offsetx[9][3]=-3  ; offsety[9][3]=1   ; offsetz[9][3]=0  ; offsetx[9][4]=-1  ; offsety[9][4]=-3  ; offsetz[9][4]=0   ; offsetx[9][5]=-1  ; offsety[9][5]=0   ; offsetz[9][5]=-3  ; offsetx[9][6]=-1  ; offsety[9][6]=0   ; offsetz[9][6]=3   ; offsetx[9][7]=-1  ; offsety[9][7]=3   ; offsetz[9][7]=0  ; offsetx[9][8]=0   ; offsety[9][8]=-3  ; offsetz[9][8]=-1  ; offsetx[9][9]=0   ; offsety[9][9]=-3  ; offsetz[9][9]=1   ; offsetx[9][10]=0   ; offsety[9][10]=-1  ; offsetz[9][10]=-3  ; offsetx[9][11]=0   ; offsety[9][11]=-1  ; offsetz[9][11]=3  ; offsetx[9][12]=0   ; offsety[9][12]=1   ; offsetz[9][12]=-3  ; offsetx[9][13]=0   ; offsety[9][13]=1   ; offsetz[9][13]=3  ; offsetx[9][14]=0   ; offsety[9][14]=3   ; offsetz[9][14]=-1  ; offsetx[9][15]=0   ; offsety[9][15]=3   ; offsetz[9][15]=1  ; offsetx[9][16]=1   ; offsety[9][16]=-3  ; offsetz[9][16]=0   ; offsetx[9][17]=1   ; offsety[9][17]=0   ; offsetz[9][17]=-3  ; offsetx[9][18]=1   ; offsety[9][18]=0   ; offsetz[9][18]=3   ; offsetx[9][19]=1   ; offsety[9][19]=3   ; offsetz[9][19]=0  ; offsetx[9][20]=3   ; offsety[9][20]=-1  ; offsetz[9][20]=0   ; offsetx[9][21]=3   ; offsety[9][21]=0   ; offsetz[9][21]=-1  ; offsetx[9][22]=3   ; offsety[9][22]=0  ; offsetz[9][22]=1   ; offsetx[9][23]=3   ; offsety[9][23]=1  ; offsetz[9][23]=0  ;
      offsetx[10][0]=-3 ; offsety[10][0]=-1 ; offsetz[10][0]=-1 ; offsetx[10][1]=-3 ; offsety[10][1]=-1 ; offsetz[10][1]=1  ; offsetx[10][2]=-3 ; offsety[10][2]=1  ; offsetz[10][2]=-1 ; offsetx[10][3]=-3 ; offsety[10][3]=1  ; offsetz[10][3]=1 ; offsetx[10][4]=-1 ; offsety[10][4]=-3 ; offsetz[10][4]=-1 ; offsetx[10][5]=-1 ; offsety[10][5]=-3 ; offsetz[10][5]=1  ; offsetx[10][6]=-1 ; offsety[10][6]=-1 ; offsetz[10][6]=-3 ; offsetx[10][7]=-1 ; offsety[10][7]=-1 ; offsetz[10][7]=3 ; offsetx[10][8]=-1 ; offsety[10][8]=1  ; offsetz[10][8]=-3 ; offsetx[10][9]=-1 ; offsety[10][9]=1  ; offsetz[10][9]=3  ; offsetx[10][10]=-1 ; offsety[10][10]=3  ; offsetz[10][10]=-1 ; offsetx[10][11]=-1 ; offsety[10][11]=3  ; offsetz[10][11]=1 ; offsetx[10][12]=1  ; offsety[10][12]=-3 ; offsetz[10][12]=-1 ; offsetx[10][13]=1  ; offsety[10][13]=-3 ; offsetz[10][13]=1 ; offsetx[10][14]=1  ; offsety[10][14]=-1 ; offsetz[10][14]=-3 ; offsetx[10][15]=1  ; offsety[10][15]=-1 ; offsetz[10][15]=3 ; offsetx[10][16]=1  ; offsety[10][16]=1  ; offsetz[10][16]=-3 ; offsetx[10][17]=1  ; offsety[10][17]=1  ; offsetz[10][17]=3  ; offsetx[10][18]=1  ; offsety[10][18]=3  ; offsetz[10][18]=-1 ; offsetx[10][19]=1  ; offsety[10][19]=3  ; offsetz[10][19]=1 ; offsetx[10][20]=3  ; offsety[10][20]=-1 ; offsetz[10][20]=-1 ; offsetx[10][21]=3  ; offsety[10][21]=-1 ; offsetz[10][21]=1  ; offsetx[10][22]=3  ; offsety[10][22]=1 ; offsetz[10][22]=-1 ; offsetx[10][23]=3  ; offsety[10][23]=1 ; offsetz[10][23]=1 ;
      offsetx[11][0]=-2 ; offsety[11][0]=-2 ; offsetz[11][0]=-2 ; offsetx[11][1]=-2 ; offsety[11][1]=-2 ; offsetz[11][1]=2  ; offsetx[11][2]=-2 ; offsety[11][2]=2  ; offsetz[11][2]=-2 ; offsetx[11][3]=-2 ; offsety[11][3]=2  ; offsetz[11][3]=2 ; offsetx[11][4]=2  ; offsety[11][4]=-2 ; offsetz[11][4]=-2 ; offsetx[11][5]=2  ; offsety[11][5]=-2 ; offsetz[11][5]=2  ; offsetx[11][6]=2  ; offsety[11][6]=2  ; offsetz[11][6]=-2 ; offsetx[11][7]=2  ; offsety[11][7]=2  ; offsetz[11][7]=2 ;
      offsetx[12][0]=-3 ; offsety[12][0]=-2 ; offsetz[12][0]=0  ; offsetx[12][1]=-3 ; offsety[12][1]=0  ; offsetz[12][1]=-2 ; offsetx[12][2]=-3 ; offsety[12][2]=0  ; offsetz[12][2]=2  ; offsetx[12][3]=-3 ; offsety[12][3]=2  ; offsetz[12][3]=0 ; offsetx[12][4]=-2 ; offsety[12][4]=-3 ; offsetz[12][4]=0  ; offsetx[12][5]=-2 ; offsety[12][5]=0  ; offsetz[12][5]=-3 ; offsetx[12][6]=-2 ; offsety[12][6]=0  ; offsetz[12][6]=3  ; offsetx[12][7]=-2 ; offsety[12][7]=3  ; offsetz[12][7]=0 ; offsetx[12][8]=0  ; offsety[12][8]=-3 ; offsetz[12][8]=-2 ; offsetx[12][9]=0  ; offsety[12][9]=-3 ; offsetz[12][9]=2  ; offsetx[12][10]=0  ; offsety[12][10]=-2 ; offsetz[12][10]=-3 ; offsetx[12][11]=0  ; offsety[12][11]=-2 ; offsetz[12][11]=3 ; offsetx[12][12]=0  ; offsety[12][12]=2  ; offsetz[12][12]=-3 ; offsetx[12][13]=0  ; offsety[12][13]=2  ; offsetz[12][13]=3 ; offsetx[12][14]=0  ; offsety[12][14]=3  ; offsetz[12][14]=-2 ; offsetx[12][15]=0  ; offsety[12][15]=3  ; offsetz[12][15]=2 ; offsetx[12][16]=2  ; offsety[12][16]=-3 ; offsetz[12][16]=0  ; offsetx[12][17]=2  ; offsety[12][17]=0  ; offsetz[12][17]=-3 ; offsetx[12][18]=2  ; offsety[12][18]=0  ; offsetz[12][18]=3  ; offsetx[12][19]=2  ; offsety[12][19]=3  ; offsetz[12][19]=0 ; offsetx[12][20]=3  ; offsety[12][20]=-2 ; offsetz[12][20]=0  ; offsetx[12][21]=3  ; offsety[12][21]=0  ; offsetz[12][21]=-2 ; offsetx[12][22]=3  ; offsety[12][22]=0 ; offsetz[12][22]=2  ; offsetx[12][23]=3  ; offsety[12][23]=2 ; offsetz[12][23]=0 ;
      offsetx[13][0]=-3 ; offsety[13][0]=-2 ; offsetz[13][0]=-1 ; offsetx[13][1]=-3 ; offsety[13][1]=-2 ; offsetz[13][1]=1  ; offsetx[13][2]=-3 ; offsety[13][2]=-1 ; offsetz[13][2]=-2 ; offsetx[13][3]=-3 ; offsety[13][3]=-1 ; offsetz[13][3]=2 ; offsetx[13][4]=-3 ; offsety[13][4]=1  ; offsetz[13][4]=-2 ; offsetx[13][5]=-3 ; offsety[13][5]=1  ; offsetz[13][5]=2  ; offsetx[13][6]=-3 ; offsety[13][6]=2  ; offsetz[13][6]=-1 ; offsetx[13][7]=-3 ; offsety[13][7]=2  ; offsetz[13][7]=1 ; offsetx[13][8]=-2 ; offsety[13][8]=-3 ; offsetz[13][8]=-1 ; offsetx[13][9]=-2 ; offsety[13][9]=-3 ; offsetz[13][9]=1  ; offsetx[13][10]=-2 ; offsety[13][10]=-1 ; offsetz[13][10]=-3 ; offsetx[13][11]=-2 ; offsety[13][11]=-1 ; offsetz[13][11]=3 ; offsetx[13][12]=-2 ; offsety[13][12]=1  ; offsetz[13][12]=-3 ; offsetx[13][13]=-2 ; offsety[13][13]=1  ; offsetz[13][13]=3 ; offsetx[13][14]=-2 ; offsety[13][14]=3  ; offsetz[13][14]=-1 ; offsetx[13][15]=-2 ; offsety[13][15]=3  ; offsetz[13][15]=1 ; offsetx[13][16]=-1 ; offsety[13][16]=-3 ; offsetz[13][16]=-2 ; offsetx[13][17]=-1 ; offsety[13][17]=-3 ; offsetz[13][17]=2  ; offsetx[13][18]=-1 ; offsety[13][18]=-2 ; offsetz[13][18]=-3 ; offsetx[13][19]=-1 ; offsety[13][19]=-2 ; offsetz[13][19]=3 ; offsetx[13][20]=-1 ; offsety[13][20]=2  ; offsetz[13][20]=-3 ; offsetx[13][21]=-1 ; offsety[13][21]=2  ; offsetz[13][21]=3  ; offsetx[13][22]=-1 ; offsety[13][22]=3 ; offsetz[13][22]=-2 ; offsetx[13][23]=-1 ; offsety[13][23]=3 ; offsetz[13][23]=2 ; offsetx[13][24]=1 ; offsety[13][24]=-3 ; offsetz[13][24]=-2 ; offsetx[13][25]=1 ; offsety[13][25]=-3 ; offsetz[13][25]=2 ; offsetx[13][26]=1 ; offsety[13][26]=-2 ; offsetz[13][26]=-3 ; offsetx[13][27]=1 ; offsety[13][27]=-2 ; offsetz[13][27]=3 ; offsetx[13][28]=1 ; offsety[13][28]=2 ; offsetz[13][28]=-3 ; offsetx[13][29]=1 ; offsety[13][29]=2 ; offsetz[13][29]=3 ; offsetx[13][30]=1 ; offsety[13][30]=3 ; offsetz[13][30]=-2 ; offsetx[13][31]=1 ; offsety[13][31]=3 ; offsetz[13][31]=2 ; offsetx[13][32]=2 ; offsety[13][32]=-3 ; offsetz[13][32]=-1 ; offsetx[13][33]=2 ; offsety[13][33]=-3 ; offsetz[13][33]=1 ; offsetx[13][34]=2 ; offsety[13][34]=-1 ; offsetz[13][34]=-3 ; offsetx[13][35]=2 ; offsety[13][35]=-1 ; offsetz[13][35]=3 ; offsetx[13][36]=2 ; offsety[13][36]=1 ; offsetz[13][36]=-3 ; offsetx[13][37]=2 ; offsety[13][37]=1 ; offsetz[13][37]=3 ; offsetx[13][38]=2 ; offsety[13][38]=3 ; offsetz[13][38]=-1 ; offsetx[13][39]=2 ; offsety[13][39]=3 ; offsetz[13][39]=1 ; offsetx[13][40]=3 ; offsety[13][40]=-2 ; offsetz[13][40]=-1 ; offsetx[13][41]=3 ; offsety[13][41]=-2 ; offsetz[13][41]=1 ; offsetx[13][42]=3 ; offsety[13][42]=-1 ; offsetz[13][42]=-2 ; offsetx[13][43]=3 ; offsety[13][43]=-1 ; offsetz[13][43]=2 ; offsetx[13][44]=3 ; offsety[13][44]=1 ; offsetz[13][44]=-2 ; offsetx[13][45]=3 ; offsety[13][45]=1 ; offsetz[13][45]=2 ; offsetx[13][46]=3 ; offsety[13][46]=2 ; offsetz[13][46]=-1 ; offsetx[13][47]=3 ; offsety[13][47]=2 ; offsetz[13][47]=1 ;
      offsetx[14][0]=-4 ; offsety[14][0]=0  ; offsetz[14][0]=0  ; offsetx[14][1]=0  ; offsety[14][1]=-4 ; offsetz[14][1]=0  ; offsetx[14][2]=0  ; offsety[14][2]=0  ; offsetz[14][2]=-4 ; offsetx[14][3]=0  ; offsety[14][3]=0  ; offsetz[14][3]=4 ; offsetx[14][4]=0  ; offsety[14][4]=4  ; offsetz[14][4]=0  ; offsetx[14][5]=4  ; offsety[14][5]=0  ; offsetz[14][5]=0  ;
      offsetx[15][0]=-3 ; offsety[15][0]=-2 ; offsetz[15][0]=-2 ; offsetx[15][1]=-3 ; offsety[15][1]=-2 ; offsetz[15][1]=2  ; offsetx[15][2]=-3 ; offsety[15][2]=2  ; offsetz[15][2]=-2 ; offsetx[15][3]=-3 ; offsety[15][3]=2  ; offsetz[15][3]=2 ; offsetx[15][4]=-2 ; offsety[15][4]=-3 ; offsetz[15][4]=-2 ; offsetx[15][5]=-2 ; offsety[15][5]=-3 ; offsetz[15][5]=2  ; offsetx[15][6]=-2 ; offsety[15][6]=-2 ; offsetz[15][6]=-3 ; offsetx[15][7]=-2 ; offsety[15][7]=-2 ; offsetz[15][7]=3 ; offsetx[15][8]=-2 ; offsety[15][8]=2  ; offsetz[15][8]=-3 ; offsetx[15][9]=-2 ; offsety[15][9]=2  ; offsetz[15][9]=3  ; offsetx[15][10]=-2 ; offsety[15][10]=3  ; offsetz[15][10]=-2 ; offsetx[15][11]=-2 ; offsety[15][11]=3  ; offsetz[15][11]=2 ; offsetx[15][12]=2  ; offsety[15][12]=-3 ; offsetz[15][12]=-2 ; offsetx[15][13]=2  ; offsety[15][13]=-3 ; offsetz[15][13]=2 ; offsetx[15][14]=2  ; offsety[15][14]=-2 ; offsetz[15][14]=-3 ; offsetx[15][15]=2  ; offsety[15][15]=-2 ; offsetz[15][15]=3 ; offsetx[15][16]=2  ; offsety[15][16]=2  ; offsetz[15][16]=-3 ; offsetx[15][17]=2  ; offsety[15][17]=2  ; offsetz[15][17]=3  ; offsetx[15][18]=2  ; offsety[15][18]=3  ; offsetz[15][18]=-2 ; offsetx[15][19]=2  ; offsety[15][19]=3  ; offsetz[15][19]=2 ; offsetx[15][20]=3  ; offsety[15][20]=-2 ; offsetz[15][20]=-2 ; offsetx[15][21]=3  ; offsety[15][21]=-2 ; offsetz[15][21]=2  ; offsetx[15][22]=3  ; offsety[15][22]=2 ; offsetz[15][22]=-2 ; offsetx[15][23]=3  ; offsety[15][23]=2 ; offsetz[15][23]=2 ;
      offsetx[16][0]=-4 ; offsety[16][0]=-1 ; offsetz[16][0]=0  ; offsetx[16][1]=-4 ; offsety[16][1]=0  ; offsetz[16][1]=-1 ; offsetx[16][2]=-4 ; offsety[16][2]=0  ; offsetz[16][2]=1  ; offsetx[16][3]=-4 ; offsety[16][3]=1  ; offsetz[16][3]=0 ; offsetx[16][4]=-1 ; offsety[16][4]=-4 ; offsetz[16][4]=0  ; offsetx[16][5]=-1 ; offsety[16][5]=0  ; offsetz[16][5]=-4 ; offsetx[16][6]=-1 ; offsety[16][6]=0  ; offsetz[16][6]=4  ; offsetx[16][7]=-1 ; offsety[16][7]=4  ; offsetz[16][7]=0 ; offsetx[16][8]=0  ; offsety[16][8]=-4 ; offsetz[16][8]=-1 ; offsetx[16][9]=0  ; offsety[16][9]=-4 ; offsetz[16][9]=1  ; offsetx[16][10]=0  ; offsety[16][10]=-1 ; offsetz[16][10]=-4 ; offsetx[16][11]=0  ; offsety[16][11]=-1 ; offsetz[16][11]=4 ; offsetx[16][12]=0  ; offsety[16][12]=1  ; offsetz[16][12]=-4 ; offsetx[16][13]=0  ; offsety[16][13]=1  ; offsetz[16][13]=4 ; offsetx[16][14]=0  ; offsety[16][14]=4  ; offsetz[16][14]=-1 ; offsetx[16][15]=0  ; offsety[16][15]=4  ; offsetz[16][15]=1 ; offsetx[16][16]=1  ; offsety[16][16]=-4 ; offsetz[16][16]=0  ; offsetx[16][17]=1  ; offsety[16][17]=0  ; offsetz[16][17]=-4 ; offsetx[16][18]=1  ; offsety[16][18]=0  ; offsetz[16][18]=4  ; offsetx[16][19]=1  ; offsety[16][19]=4  ; offsetz[16][19]=0 ; offsetx[16][20]=4  ; offsety[16][20]=-1 ; offsetz[16][20]=0  ; offsetx[16][21]=4  ; offsety[16][21]=0  ; offsetz[16][21]=-1 ; offsetx[16][22]=4  ; offsety[16][22]=0 ; offsetz[16][22]=1  ; offsetx[16][23]=4  ; offsety[16][23]=1 ; offsetz[16][23]=0 ;
      offsetx[17][0]=-3 ; offsety[17][0]=-3 ; offsetz[17][0]=0  ; offsetx[17][1]=-3 ; offsety[17][1]=0  ; offsetz[17][1]=-3 ; offsetx[17][2]=-3 ; offsety[17][2]=0  ; offsetz[17][2]=3  ; offsetx[17][3]=-3 ; offsety[17][3]=3  ; offsetz[17][3]=0 ; offsetx[17][4]=0  ; offsety[17][4]=-3 ; offsetz[17][4]=-3 ; offsetx[17][5]=0  ; offsety[17][5]=-3 ; offsetz[17][5]=3  ; offsetx[17][6]=0  ; offsety[17][6]=3  ; offsetz[17][6]=-3 ; offsetx[17][7]=0  ; offsety[17][7]=3  ; offsetz[17][7]=3 ; offsetx[17][8]=3  ; offsety[17][8]=-3 ; offsetz[17][8]=0  ; offsetx[17][9]=3  ; offsety[17][9]=0  ; offsetz[17][9]=-3 ; offsetx[17][10]=3  ; offsety[17][10]=0  ; offsetz[17][10]=3  ; offsetx[17][11]=3  ; offsety[17][11]=3  ; offsetz[17][11]=0 ;
      offsetx[18][0]=-4 ; offsety[18][0]=-1 ; offsetz[18][0]=-1 ; offsetx[18][1]=-4 ; offsety[18][1]=-1 ; offsetz[18][1]=1  ; offsetx[18][2]=-4 ; offsety[18][2]=1  ; offsetz[18][2]=-1 ; offsetx[18][3]=-4 ; offsety[18][3]=1  ; offsetz[18][3]=1 ; offsetx[18][4]=-1 ; offsety[18][4]=-4 ; offsetz[18][4]=-1 ; offsetx[18][5]=-1 ; offsety[18][5]=-4 ; offsetz[18][5]=1  ; offsetx[18][6]=-1 ; offsety[18][6]=-1 ; offsetz[18][6]=-4 ; offsetx[18][7]=-1 ; offsety[18][7]=-1 ; offsetz[18][7]=4 ; offsetx[18][8]=-1 ; offsety[18][8]=1  ; offsetz[18][8]=-4 ; offsetx[18][9]=-1 ; offsety[18][9]=1  ; offsetz[18][9]=4  ; offsetx[18][10]=-1 ; offsety[18][10]=4  ; offsetz[18][10]=-1 ; offsetx[18][11]=-1 ; offsety[18][11]=4  ; offsetz[18][11]=1 ; offsetx[18][12]=1  ; offsety[18][12]=-4 ; offsetz[18][12]=-1 ; offsetx[18][13]=1  ; offsety[18][13]=-4 ; offsetz[18][13]=1 ; offsetx[18][14]=1  ; offsety[18][14]=-1 ; offsetz[18][14]=-4 ; offsetx[18][15]=1  ; offsety[18][15]=-1 ; offsetz[18][15]=4 ; offsetx[18][16]=1  ; offsety[18][16]=1  ; offsetz[18][16]=-4 ; offsetx[18][17]=1  ; offsety[18][17]=1  ; offsetz[18][17]=4  ; offsetx[18][18]=1  ; offsety[18][18]=4  ; offsetz[18][18]=-1 ; offsetx[18][19]=1  ; offsety[18][19]=4  ; offsetz[18][19]=1 ; offsetx[18][20]=4  ; offsety[18][20]=-1 ; offsetz[18][20]=-1 ; offsetx[18][21]=4  ; offsety[18][21]=-1 ; offsetz[18][21]=1  ; offsetx[18][22]=4  ; offsety[18][22]=1 ; offsetz[18][22]=-1 ; offsetx[18][23]=4  ; offsety[18][23]=1 ; offsetz[18][23]=1 ;
      offsetx[19][0]=-3 ; offsety[19][0]=-3 ; offsetz[19][0]=-1 ; offsetx[19][1]=-3 ; offsety[19][1]=-3 ; offsetz[19][1]=1  ; offsetx[19][2]=-3 ; offsety[19][2]=-1 ; offsetz[19][2]=-3 ; offsetx[19][3]=-3 ; offsety[19][3]=-1 ; offsetz[19][3]=3 ; offsetx[19][4]=-3 ; offsety[19][4]=1  ; offsetz[19][4]=-3 ; offsetx[19][5]=-3 ; offsety[19][5]=1  ; offsetz[19][5]=3  ; offsetx[19][6]=-3 ; offsety[19][6]=3  ; offsetz[19][6]=-1 ; offsetx[19][7]=-3 ; offsety[19][7]=3  ; offsetz[19][7]=1 ; offsetx[19][8]=-1 ; offsety[19][8]=-3 ; offsetz[19][8]=-3 ; offsetx[19][9]=-1 ; offsety[19][9]=-3 ; offsetz[19][9]=3  ; offsetx[19][10]=-1 ; offsety[19][10]=3  ; offsetz[19][10]=-3 ; offsetx[19][11]=-1 ; offsety[19][11]=3  ; offsetz[19][11]=3 ; offsetx[19][12]=1  ; offsety[19][12]=-3 ; offsetz[19][12]=-3 ; offsetx[19][13]=1  ; offsety[19][13]=-3 ; offsetz[19][13]=3 ; offsetx[19][14]=1  ; offsety[19][14]=3  ; offsetz[19][14]=-3 ; offsetx[19][15]=1  ; offsety[19][15]=3  ; offsetz[19][15]=3 ; offsetx[19][16]=3  ; offsety[19][16]=-3 ; offsetz[19][16]=-1 ; offsetx[19][17]=3  ; offsety[19][17]=-3 ; offsetz[19][17]=1  ; offsetx[19][18]=3  ; offsety[19][18]=-1 ; offsetz[19][18]=-3 ; offsetx[19][19]=3  ; offsety[19][19]=-1 ; offsetz[19][19]=3 ; offsetx[19][20]=3  ; offsety[19][20]=1  ; offsetz[19][20]=-3 ; offsetx[19][21]=3  ; offsety[19][21]=1  ; offsetz[19][21]=3  ; offsetx[19][22]=3  ; offsety[19][22]=3 ; offsetz[19][22]=-1 ; offsetx[19][23]=3  ; offsety[19][23]=3 ; offsetz[19][23]=1 ;
      offsetx[20][0]=-4 ; offsety[20][0]=-2 ; offsetz[20][0]=0  ; offsetx[20][1]=-4 ; offsety[20][1]=0  ; offsetz[20][1]=-2 ; offsetx[20][2]=-4 ; offsety[20][2]=0  ; offsetz[20][2]=2  ; offsetx[20][3]=-4 ; offsety[20][3]=2  ; offsetz[20][3]=0 ; offsetx[20][4]=-2 ; offsety[20][4]=-4 ; offsetz[20][4]=0  ; offsetx[20][5]=-2 ; offsety[20][5]=0  ; offsetz[20][5]=-4 ; offsetx[20][6]=-2 ; offsety[20][6]=0  ; offsetz[20][6]=4  ; offsetx[20][7]=-2 ; offsety[20][7]=4  ; offsetz[20][7]=0 ; offsetx[20][8]=0  ; offsety[20][8]=-4 ; offsetz[20][8]=-2 ; offsetx[20][9]=0  ; offsety[20][9]=-4 ; offsetz[20][9]=2  ; offsetx[20][10]=0  ; offsety[20][10]=-2 ; offsetz[20][10]=-4 ; offsetx[20][11]=0  ; offsety[20][11]=-2 ; offsetz[20][11]=4 ; offsetx[20][12]=0  ; offsety[20][12]=2  ; offsetz[20][12]=-4 ; offsetx[20][13]=0  ; offsety[20][13]=2  ; offsetz[20][13]=4 ; offsetx[20][14]=0  ; offsety[20][14]=4  ; offsetz[20][14]=-2 ; offsetx[20][15]=0  ; offsety[20][15]=4  ; offsetz[20][15]=2 ; offsetx[20][16]=2  ; offsety[20][16]=-4 ; offsetz[20][16]=0  ; offsetx[20][17]=2  ; offsety[20][17]=0  ; offsetz[20][17]=-4 ; offsetx[20][18]=2  ; offsety[20][18]=0  ; offsetz[20][18]=4  ; offsetx[20][19]=2  ; offsety[20][19]=4  ; offsetz[20][19]=0 ; offsetx[20][20]=4  ; offsety[20][20]=-2 ; offsetz[20][20]=0  ; offsetx[20][21]=4  ; offsety[20][21]=0  ; offsetz[20][21]=-2 ; offsetx[20][22]=4  ; offsety[20][22]=0 ; offsetz[20][22]=2  ; offsetx[20][23]=4  ; offsety[20][23]=2 ; offsetz[20][23]=0 ;
      offsetx[21][0]=-4 ; offsety[21][0]=-2 ; offsetz[21][0]=-1 ; offsetx[21][1]=-4 ; offsety[21][1]=-2 ; offsetz[21][1]=1  ; offsetx[21][2]=-4 ; offsety[21][2]=-1 ; offsetz[21][2]=-2 ; offsetx[21][3]=-4 ; offsety[21][3]=-1 ; offsetz[21][3]=2 ; offsetx[21][4]=-4 ; offsety[21][4]=1  ; offsetz[21][4]=-2 ; offsetx[21][5]=-4 ; offsety[21][5]=1  ; offsetz[21][5]=2  ; offsetx[21][6]=-4 ; offsety[21][6]=2  ; offsetz[21][6]=-1 ; offsetx[21][7]=-4 ; offsety[21][7]=2  ; offsetz[21][7]=1 ; offsetx[21][8]=-2 ; offsety[21][8]=-4 ; offsetz[21][8]=-1 ; offsetx[21][9]=-2 ; offsety[21][9]=-4 ; offsetz[21][9]=1  ; offsetx[21][10]=-2 ; offsety[21][10]=-1 ; offsetz[21][10]=-4 ; offsetx[21][11]=-2 ; offsety[21][11]=-1 ; offsetz[21][11]=4 ; offsetx[21][12]=-2 ; offsety[21][12]=1  ; offsetz[21][12]=-4 ; offsetx[21][13]=-2 ; offsety[21][13]=1  ; offsetz[21][13]=4 ; offsetx[21][14]=-2 ; offsety[21][14]=4  ; offsetz[21][14]=-1 ; offsetx[21][15]=-2 ; offsety[21][15]=4  ; offsetz[21][15]=1 ; offsetx[21][16]=-1 ; offsety[21][16]=-4 ; offsetz[21][16]=-2 ; offsetx[21][17]=-1 ; offsety[21][17]=-4 ; offsetz[21][17]=2  ; offsetx[21][18]=-1 ; offsety[21][18]=-2 ; offsetz[21][18]=-4 ; offsetx[21][19]=-1 ; offsety[21][19]=-2 ; offsetz[21][19]=4 ; offsetx[21][20]=-1 ; offsety[21][20]=2  ; offsetz[21][20]=-4 ; offsetx[21][21]=-1 ; offsety[21][21]=2  ; offsetz[21][21]=4  ; offsetx[21][22]=-1 ; offsety[21][22]=4 ; offsetz[21][22]=-2 ; offsetx[21][23]=-1 ; offsety[21][23]=4 ; offsetz[21][23]=2 ; offsetx[21][24]=1 ; offsety[21][24]=-4 ; offsetz[21][24]=-2 ; offsetx[21][25]=1 ; offsety[21][25]=-4 ; offsetz[21][25]=2 ; offsetx[21][26]=1 ; offsety[21][26]=-2 ; offsetz[21][26]=-4 ; offsetx[21][27]=1 ; offsety[21][27]=-2 ; offsetz[21][27]=4 ; offsetx[21][28]=1 ; offsety[21][28]=2 ; offsetz[21][28]=-4 ; offsetx[21][29]=1 ; offsety[21][29]=2 ; offsetz[21][29]=4 ; offsetx[21][30]=1 ; offsety[21][30]=4 ; offsetz[21][30]=-2 ; offsetx[21][31]=1 ; offsety[21][31]=4 ; offsetz[21][31]=2 ; offsetx[21][32]=2 ; offsety[21][32]=-4 ; offsetz[21][32]=-1 ; offsetx[21][33]=2 ; offsety[21][33]=-4 ; offsetz[21][33]=1 ; offsetx[21][34]=2 ; offsety[21][34]=-1 ; offsetz[21][34]=-4 ; offsetx[21][35]=2 ; offsety[21][35]=-1 ; offsetz[21][35]=4 ; offsetx[21][36]=2 ; offsety[21][36]=1 ; offsetz[21][36]=-4 ; offsetx[21][37]=2 ; offsety[21][37]=1 ; offsetz[21][37]=4 ; offsetx[21][38]=2 ; offsety[21][38]=4 ; offsetz[21][38]=-1 ; offsetx[21][39]=2 ; offsety[21][39]=4 ; offsetz[21][39]=1 ; offsetx[21][40]=4 ; offsety[21][40]=-2 ; offsetz[21][40]=-1 ; offsetx[21][41]=4 ; offsety[21][41]=-2 ; offsetz[21][41]=1 ; offsetx[21][42]=4 ; offsety[21][42]=-1 ; offsetz[21][42]=-2 ; offsetx[21][43]=4 ; offsety[21][43]=-1 ; offsetz[21][43]=2 ; offsetx[21][44]=4 ; offsety[21][44]=1 ; offsetz[21][44]=-2 ; offsetx[21][45]=4 ; offsety[21][45]=1 ; offsetz[21][45]=2 ; offsetx[21][46]=4 ; offsety[21][46]=2 ; offsetz[21][46]=-1 ; offsetx[21][47]=4 ; offsety[21][47]=2 ; offsetz[21][47]=1 ;
      offsetx[22][0]=-3 ; offsety[22][0]=-3 ; offsetz[22][0]=-2 ; offsetx[22][1]=-3 ; offsety[22][1]=-3 ; offsetz[22][1]=2  ; offsetx[22][2]=-3 ; offsety[22][2]=-2 ; offsetz[22][2]=-3 ; offsetx[22][3]=-3 ; offsety[22][3]=-2 ; offsetz[22][3]=3 ; offsetx[22][4]=-3 ; offsety[22][4]=2  ; offsetz[22][4]=-3 ; offsetx[22][5]=-3 ; offsety[22][5]=2  ; offsetz[22][5]=3  ; offsetx[22][6]=-3 ; offsety[22][6]=3  ; offsetz[22][6]=-2 ; offsetx[22][7]=-3 ; offsety[22][7]=3  ; offsetz[22][7]=2 ; offsetx[22][8]=-2 ; offsety[22][8]=-3 ; offsetz[22][8]=-3 ; offsetx[22][9]=-2 ; offsety[22][9]=-3 ; offsetz[22][9]=3  ; offsetx[22][10]=-2 ; offsety[22][10]=3  ; offsetz[22][10]=-3 ; offsetx[22][11]=-2 ; offsety[22][11]=3  ; offsetz[22][11]=3 ; offsetx[22][12]=2  ; offsety[22][12]=-3 ; offsetz[22][12]=-3 ; offsetx[22][13]=2  ; offsety[22][13]=-3 ; offsetz[22][13]=3 ; offsetx[22][14]=2  ; offsety[22][14]=3  ; offsetz[22][14]=-3 ; offsetx[22][15]=2  ; offsety[22][15]=3  ; offsetz[22][15]=3 ; offsetx[22][16]=3  ; offsety[22][16]=-3 ; offsetz[22][16]=-2 ; offsetx[22][17]=3  ; offsety[22][17]=-3 ; offsetz[22][17]=2  ; offsetx[22][18]=3  ; offsety[22][18]=-2 ; offsetz[22][18]=-3 ; offsetx[22][19]=3  ; offsety[22][19]=-2 ; offsetz[22][19]=3 ; offsetx[22][20]=3  ; offsety[22][20]=2  ; offsetz[22][20]=-3 ; offsetx[22][21]=3  ; offsety[22][21]=2  ; offsetz[22][21]=3  ; offsetx[22][22]=3  ; offsety[22][22]=3 ; offsetz[22][22]=-2 ; offsetx[22][23]=3  ; offsety[22][23]=3 ; offsetz[22][23]=2 ;
   }
   int ii=0;

   //leggy and kernelB, load coefficients
   if (leggy==1 && useKernelB) {
      for (ii=0;ii<maxNSh;ii++){
         gammaK_h[ii]=gammaP[ii];
      }
   }
   //generic, kernelA
   else {
      //origin
      gammaK_h[0]=gammaP[0];
      offset_h[0]=0;

      ii=1;
      //iterate through shells and vectors, load linear offsets and coefficients
      for(int ss=0;ss<NSh;ss++){
         for(int qq=0;qq<NN[ss];qq++){
            offset_h[ii]=offsetx[ss][qq]+offsety[ss][qq]*Nxh+offsetz[ss][qq]*area;
            gammaK_h[ii]=gammaP[ss+1];
            ii++;
         }
      }
   }

   //-----------------------------------------------
   // Display settings
   //-----------------------------------------------
   printf("Grid dims   : %d x %d x %d = %d\n",Nxh,Nyh,Nzh,Nxh*Nyh*Nzh);
   printf("In-grid dims: %d x %d x %d = %d\n",Nx,Ny,Nz,Nx*Ny*Nz);
   printf("Memory usage: %.2f GiB\n",1.0*Nxh*Nyh*Nzh*2.0*sizeof(ReaL)/1073741824.0);
   printf("Duration    : %.d\n",N);
   printf("Type        : %d\n",stype);
   printf("Leggy       : %d\n",leggy);
   printf("useKernelB  : %d\n",useKernelB);
   printf("Nh          : %d\n",Nh);
   printf("Precision   : Single\n");

   //-----------------------------------------------
   // Constant memory
   //-----------------------------------------------
   cuErr( cudaMemcpyToSymbol(gammaK,gammaK_h,sizeof(ReaL)*(K)) );
   cuErr( cudaMemcpyToSymbol(offset,offset_h,sizeof(int)*K) );

   // ------------------------------------------
   // Set up grid and blocks
   //-------------------------------------------
   int Gx          = (Nx-1)/Bx+1;
   int Gy          = (Ny-1)/By+1;
   int Gz          = (Nz-1)/Bz+1;

   dim3 dimBlockInt(Bx, By, Bz);
   dim3 dimGridInt(Gx, Gy, Gz);
   dim3 dimBlockIO(1, 1, 1);
   dim3 dimGridIO(1, 1, 1);

   size_t pr_size  = sizeof(ReaL);
   size_t mem_size = Nxh*Nyh*Nzh*pr_size;
   ReaL *out_d, *u0_d, *u1_d, *dummy_ptr;
   int n;

   //-------------------------------------------
   // initialise memory on host
   //-------------------------------------------
   ReaL *out_h  = (ReaL *)calloc(N,pr_size);
   unsigned long int vol  = (unsigned long int)Nxh*Nyh*Nzh;

   ReaL *u0_h, *u1_h;
   u0_h   = (ReaL *)calloc(vol,sizeof(ReaL));
   u1_h  = (ReaL *)calloc(vol,sizeof(ReaL));
   if((out_h == NULL) || (u0_h == NULL) || (u1_h == NULL)){
      printf("\nmemory alloc failed...\n");
      exit(EXIT_FAILURE);
   }
   // ---------------------------------------------------------
   // initial conditions
   // ---------------------------------------------------------
   int cp=0;
   for(int k=0;k<Nzh;k++){
      for(int j=0;j<Nyh;j++){
         for(int i=0;i<Nxh;i++){
            cp  = k*area+(j*Nxh+i);
            u0_h[cp]=randomFloat();
            u1_h[cp]=randomFloat();
         }
      }
   }

   printf("Block       : %d x %d x %d = %d\n",Bx,By,Bz,Bx*By*Bz);
   printf("Grid        : %d x %d x %d = %d\n",Gx,Gy,Gz,Gx*Gy*Gz);
   //-------------------------------------------
   // Initialise memory on device
   //-------------------------------------------
   cuErr( cudaMalloc(&u0_d, mem_size) );    cuErr( cudaMemcpy(u0_d, u0_h, mem_size,cudaMemcpyHostToDevice) );
   cuErr( cudaMalloc(&u1_d, mem_size) );    cuErr( cudaMemcpy(u1_d, u1_h, mem_size,cudaMemcpyHostToDevice) );
   cuErr( cudaMalloc(&out_d, N*pr_size) );  cuErr( cudaMemset(out_d, 0, N*pr_size) );

   cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);

   printf("\nGrab a cocktail, running the scheme...\n");
   // ---------------------------------------------------------
   // Time loop
   // ---------------------------------------------------------

   time_t start = time(NULL);

   for(n=0;n<N;n++)
   {
      if (useKernelB==1 && leggy==1)   KernelB<<<dimGridInt,dimBlockInt>>>(u0_d,u1_d);
      else                             KernelA<<<dimGridInt,dimBlockInt>>>(u0_d,u1_d);

      //read from output location
      oout<<<dimGridIO,dimBlockIO>>>(u0_d,out_d,n);

      //time-stepping pointer swap
      dummy_ptr = u1_d;
      u1_d  = u0_d;
      u0_d   = dummy_ptr;
   }

   //-------------------------------------------
   // print process time
   //-------------------------------------------
   checkLastCUDAError("Kernel");
   cuErr( cudaDeviceSynchronize() );
   time_t end = time(NULL);
   printf("\nProcess time : %ld seconds\n", (end-start) );

   //-------------------------------------------
   // copy result back from device
   //-------------------------------------------
   cuErr( cudaMemcpy(out_h, out_d, N*pr_size, cudaMemcpyDeviceToHost) );
   //printLastSamples(out_h, N, 5);

   //-------------------------------------------
   // Free memory
   //-------------------------------------------
   cudaFree(u0_d);cudaFree(u1_d);

   cudaDeviceReset();
   printf("\nPut down that cocktail... Simulation complete.\n\n");

}

__global__ void oout(ReaL *u0,ReaL *out,int n)
{
   // Read out
   out[n] = u0[(Rz*area)+(Ry*Nxh+Rx)];
}

//generic kernel
__global__ void KernelA(ReaL *u0, const ReaL * __restrict__ u1)
{
   // get ix,iy,iz from thread and block Id's
   int ix = blockIdx.x * Bx + threadIdx.x + Nh;
   int iy = blockIdx.y * By + threadIdx.y + Nh;
   int iz = blockIdx.z * Bz + threadIdx.z + Nh;
   //condition to prevent illegal memory accesses
   if (ix<(Nxh-Nh) && iy<(Nyh-Nh) && iz<(Nzh-Nh)) {
         //get linear index of current point
         int cp  = iz*area+(iy*Nxh+ix);
         //read previous value from global memory
         ReaL tmp=-u0[cp];
         //general stencil operation
         for(int k=0;k<K;k++){
               tmp += gammaK[k]*u1[cp+offset[k]];
         }
         //write final value back to global memory
         u0[cp] = tmp;
   }
}

//leggy kernel
__global__ void KernelB(ReaL *u0, const ReaL * __restrict__ u1)
{
   // get ix,iy,iz from thread and block Id's
   int ix = blockIdx.x * Bx + threadIdx.x + Nh;
   int iy = blockIdx.y * By + threadIdx.y + Nh;
   int iz = blockIdx.z * Bz + threadIdx.z + Nh;
   //condition to prevent illegal memory accesses
   if (ix<(Nxh-Nh) && iy<(Nyh-Nh) && iz<(Nzh-Nh)) {
         //get linear index of current point
         int cp  = iz*area+(iy*Nxh+ix);
         //read previous value from global memory
         ReaL tmp=-u0[cp];
         //leggy stencil operation
         for(int m=-Nh;m<=Nh;m++){
                  ReaL gammaK_m=gammaK[abs(m)];
                  //u1 is read through texture binding
                  tmp += gammaK_m*u1[cp-m];
                  tmp += gammaK_m*u1[cp-m*Nxh];
                  tmp += gammaK_m*u1[cp-m*area];
         }
         //write final value back to global memory
         u0[cp] = tmp;
   }
}

